import itertools
import textwrap
from flippy.transforms import *
from flippy.interpreter import CPSInterpreter
from flippy.hashable import hashablelist, hashabledict
from flippy import keep_deterministic

def trampoline(thunk):
    while callable(thunk):
        thunk = thunk()
    return thunk

def interpret(func, *args, **kwargs):
    interpreter = CPSInterpreter()
    context = {
        **func.__globals__,
        **interpreter.get_closure(func),
        "_cps": interpreter,
        "hashablelist": hashablelist,
        "hashabledict": hashabledict,
    }
    print(func.__name__, context)
    code = interpreter.transform_from_func(func)
    print(code)
    exec(ast.unparse(code), context)
    trans_func = context[func.__name__]
    return trampoline(trans_func(*args, **kwargs))

class ScopeTools:
    @classmethod
    def pack(cls, lineno, names, *, hashabledict=False):
        constructor = 'hashabledict' if hashabledict else ''
        import_ = 'from flippy.hashable import hashabledict' if hashabledict else ''
        return ast.parse(textwrap.dedent(f'''
            {import_}
            _locals_{lineno} = locals()
            _scope_{lineno} = {constructor}({{name: _locals_{lineno}[name] for name in {names} if name in _locals_{lineno}}})
        ''')).body
    @classmethod
    def unpack(cls, lineno, names):
        return ast.parse(textwrap.dedent('\n'.join(f'''
            if '{n}' in _scope_{lineno}:
                {n} = _scope_{lineno}['{n}']
        ''' for n in names))).body
    @classmethod
    def pack_unpack(cls, lineno, names, *, tag=None, hashabledict=False):
        if tag is None:
            tag = f'_{lineno}'
        return {
            f'pack{tag}': cls.pack(lineno, names, hashabledict=hashabledict),
            f'unpack{tag}': cls.unpack(lineno, names),
        }
    def fill(code, replacements):
        return ast.unparse(Placeholder.fill(ast.parse(textwrap.dedent(code)), replacements))

def helper_in_module(x):
    return x ** 2

def main_in_module_helper_in_module():
    return helper_in_module(3)

def main_in_module_helper_in_closure():
    def helper_in_closure(x):
        return x ** 2
    return helper_in_closure(3)

def test_x():
    def helper_in_function(x):
        return x ** 2
    def main_in_function_helper_in_function():
        return helper_in_function(3)

    def main_in_function_helper_in_closure():
        def helper_in_closure(x):
            return x ** 2
        return helper_in_closure(3)

    assert interpret(main_in_module_helper_in_closure) == 9
    assert interpret(main_in_module_helper_in_module) == 9
    assert interpret(main_in_function_helper_in_closure) == 9
    assert interpret(main_in_function_helper_in_function) == 9

def test_desugaring_transform():
    src_compiled = [
        ("b = f(g(a))", "b = f(g(a))"),
        (
            "a = lambda x: g(h(x))",
            textwrap.dedent("""
            def __v0(x):
                return g(h(x))
            a = __v0
            """)
        ),
        (
            "a = g(h(x)) if f(x) else h(g(x))",
            textwrap.dedent("""
            def __v0():
                __v1 = f(x)
                if __v1:
                    __v2 = g(h(x))
                else:
                    __v2 = h(g(x))
                return __v2
            a = __v0()
            """)
        ),
        (
            "d = (lambda x, y=1: g(x)*100)()",
            textwrap.dedent("""
            def __v0(x, y=1):
                return g(x)*100
            d = __v0()
            """)
        ),
        (
            textwrap.dedent("""
            def f():
                pass
            """),
            textwrap.dedent("""
            def f():
                pass
                return None
            """)
        ),

        # Various cases for assignment
        ('x += 123', 'x = x + 123'),
        ('x.y += 123', 'x.y = x.y + 123'),
        ('x[0] += 123', 'x[0] = x[0] + 123'),
        ('x[0].prop += 123', 'x[0].prop = x[0].prop + 123'),
        ('x: int = 123', 'x = 123'),
        ('x: int', ''),

        # Loops
        (
            textwrap.dedent('''
            while a:
                b
            '''),
            textwrap.dedent('''
            while True:
                if not a:
                    break
                b
            '''),
        ),
        (
            textwrap.dedent('''
            for x in range(3):
                x
            '''),
            textwrap.dedent('''
            _for_iter_2 = range(3)
            _for_idx_2 = 0
            while True:
                if not _for_idx_2 < len(_for_iter_2):
                    break
                x = _for_iter_2[_for_idx_2]
                _for_idx_2 = _for_idx_2 + 1
                x
            '''),
        ),

        # List Comprehensions
        (
            '[x for x in range(3)]',
            textwrap.dedent('''
            from flippy import recursive_reduce
            def __v0(__acc, __target):
                x = __target
                return __acc + [x]
            recursive_reduce(__v0, range(3), [])
            '''),
        ),
        (
            # Only one test
            '[x for x in range(3) if x]',
            textwrap.dedent('''
            from flippy import recursive_reduce
            def __v0(__acc, __target):
                x = __target
                def __v1():
                    __v2 = x
                    if __v2:
                        __v3 = __acc + [x]
                    else:
                        __v3 = __acc
                    return __v3
                return __v1()
            recursive_reduce(__v0, range(3), [])
            '''),
        ),
        (
            # Two tests
            '[x for x in range(3) if x if x**2]',
            textwrap.dedent('''
            from flippy import recursive_reduce
            def __v0(__acc, __target):
                x = __target
                def __v1():
                    __v4 = x
                    if __v4:
                        __v5 = x ** 2
                    else:
                        __v5 = __v4
                    __v2 = __v5
                    if __v2:
                        __v3 = __acc + [x]
                    else:
                        __v3 = __acc
                    return __v3
                return __v1()
            recursive_reduce(__v0, range(3), [])
            '''),
        ),
        (
            '[(x, y) for x in range(3) if x for y in range(4)]',
            textwrap.dedent('''
            from flippy import recursive_reduce
            def __v1(__acc, __target):
                x = __target
                def __v0(__acc, __target):
                    y = __target
                    return __acc + [(x, y)]
                def __v2():
                    __v3 = x
                    if __v3:
                        __v4 = __acc + recursive_reduce(__v0, range(4), [])
                    else:
                        __v4 = __acc
                    return __v4
                return __v2()
            recursive_reduce(__v1, range(3), [])
            '''),
        ),

        # Set/Dict Comprehensions
        (
            '{x for x in range(3)}',
            textwrap.dedent('''
            from flippy import recursive_reduce
            def __v0(__acc, __target):
                x = __target
                return __acc | {x}
            recursive_reduce(__v0, range(3), set())
            '''),
        ),
        (
            '{x: x**2 for x in range(3)}',
            textwrap.dedent('''
            from flippy import recursive_reduce
            def __v0(__acc, __target):
                x = __target
                return __acc | {x: x**2}
            recursive_reduce(__v0, range(3), {})
            '''),
        ),
        (
            '{x: y**2 for x, y in {}.items()}',
            textwrap.dedent('''
            from flippy import recursive_reduce
            def __v0(__acc, __target):
                (x, y) = __target
                return __acc | {x: y**2}
            recursive_reduce(__v0, {}.items(), {})
            '''),
        ),

    ]

    for src, comp in src_compiled:
        node = ast.parse(src)
        node = DesugaringTransform()(node)
        assert_compare_ast(node, ast.parse(comp))

def compare_sourcecode_to_equivalent_sourcecode(src, exp_src):
    node = ast.parse(src)
    targ_node = ast.parse(exp_src)
    node = DesugaringTransform()(node)
    targ_node = DesugaringTransform()(targ_node)
    assert_compare_ast(targ_node, node)

    src_context = {}
    exec(src, src_context)
    exp_context = {}
    exec(exp_src, exp_context)
    for args in [(1, 2, 3), ('b', 'a', ''), (True, False, True)]:
        assert src_context['f'](*args) == exp_context['f'](*args)

def test_multiple_and_transform():
    src = textwrap.dedent("""
    def f(a, b, c):
        return a and b and c
    """)
    exp_src = textwrap.dedent("""
    def f(a, b, c):
        return ((a and b) and c)
    """)
    compare_sourcecode_to_equivalent_sourcecode(src, exp_src)

def test_multiple_or_transform():
    src = textwrap.dedent("""
    def f(a, b, c):
        return a or b or c
    """)
    exp_src = textwrap.dedent("""
    def f(a, b, c):
        return ((a or b) or c)
    """)
    compare_sourcecode_to_equivalent_sourcecode(src, exp_src)

def test_cps_call():
    # Basic case of CPS.
    check_cps_transform('''
    def fn(x):
        return x
    ''', '''
    @lambda fn: CPSFunction(fn, 'def fn(x):\\n    return x')
    def fn(x, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = 'def fn(x):\\n    return x'
        return lambda : _cont(x)
    ''', check_args=[('a',)])

    # Basic test of names defined by function arguments + assignment
    check_cps_transform('''
    def fn(x):
        z = 0
        y = sum([1, 2, 3])
        x = x + 1
        z = z + 1
        return x + y + z
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, 'def fn(x):\\n    z = 0\\n    y = sum([1, 2, 3])\\n    x = x + 1\\n    z = z + 1\\n    return x + y + z')
    def fn(x, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = 'def fn(x):\\n    z = 0\\n    y = sum([1, 2, 3])\\n    x = x + 1\\n    z = z + 1\\n    return x + y + z'
        z = 0
        {Placeholder.new('pack_0')}

        def _cont_0(_res_0):
            {Placeholder.new('unpack_0')}

            y = _res_0
            x = x + 1
            z = z + 1
            return lambda : _cont(x + y + z)
        return lambda : _cps.interpret(sum, cont=_cont_0, stack=_stack, func_src=__func_src, locals_=_locals_0, call_id=0, lineno=2)([1, 2, 3])
    ''', ScopeTools.pack_unpack(0, ['x', 'y', 'z'])), check_args=[(0,), (1,), (2,)])

    # Making sure things still work well in nested continuations.
    check_cps_transform('''
    def fn(y):
        y = sum([y, 1])
        y = sum([y, 2])
        return y
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, 'def fn(y):\\n    y = sum([y, 1])\\n    y = sum([y, 2])\\n    return y')
    def fn(y, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = 'def fn(y):\\n    y = sum([y, 1])\\n    y = sum([y, 2])\\n    return y'
        {Placeholder.new('pack_0')}

        def _cont_0(_res_0):
            {Placeholder.new('unpack_0')}
            y = _res_0
            {Placeholder.new('pack_1')}

            def _cont_1(_res_1):
                {Placeholder.new('unpack_1')}
                y = _res_1
                return lambda : _cont(y)
            return lambda : _cps.interpret(sum, cont=_cont_1, stack=_stack, func_src=__func_src, locals_=_locals_1, call_id=1, lineno=2)([y, 2])
        return lambda : _cps.interpret(sum, cont=_cont_0, stack=_stack, func_src=__func_src, locals_=_locals_0, call_id=0, lineno=1)([y, 1])
    ''', {
        **ScopeTools.pack_unpack(0, ['y']),
        **ScopeTools.pack_unpack(1, ['y']),
    }), check_args=[(0,), (1,)])

    # Testing destructuring.
    check_cps_transform('''
    def fn(x):
        [y, z] = x
        sum([])
        return y + z
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, 'def fn(x):\\n    [y, z] = x\\n    sum([])\\n    return y + z')
    def fn(x, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = 'def fn(x):\\n    [y, z] = x\\n    sum([])\\n    return y + z'
        [y, z] = x
        {Placeholder.new('pack_0')}

        def _cont_0(_res_0):
            {Placeholder.new('unpack_0')}
            _res_0
            return lambda : _cont(y + z)
        return lambda : _cps.interpret(sum, cont=_cont_0, stack=_stack, func_src=__func_src, locals_=_locals_0, call_id=0, lineno=2)([])
    ''', {
        **ScopeTools.pack_unpack(0, ['x', 'y', 'z']),
    }), check_args=[([1, 2],), ([7, 3],)])

def test_cps_new_vars_scope_packing():
    # checking that new variables are added at the appropriate time
    check_cps_transform('''
    def fn():
        x = 3
        sum([])
        y = 4
        sum([])
        z = 5
        return x + y + z
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, 'def fn():\\n    x = 3\\n    sum([])\\n    y = 4\\n    sum([])\\n    z = 5\\n    return x + y + z')
    def fn(*, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = 'def fn():\\n    x = 3\\n    sum([])\\n    y = 4\\n    sum([])\\n    z = 5\\n    return x + y + z'
        x = 3
        {Placeholder.new('pack_0')}

        def _cont_0(_res_0):
            {Placeholder.new('unpack_0')}
            _res_0
            y = 4
            {Placeholder.new('pack_1')}

            def _cont_1(_res_1):
                {Placeholder.new('unpack_1')}
                _res_1
                z = 5
                return lambda : _cont(x + y + z)
            return lambda : _cps.interpret(sum, cont=_cont_1, stack=_stack, func_src=__func_src, locals_=_locals_1, call_id=1, lineno=4)([])
        return lambda : _cps.interpret(sum, cont=_cont_0, stack=_stack, func_src=__func_src, locals_=_locals_0, call_id=0, lineno=2)([])
    ''', {
        **ScopeTools.pack_unpack(0, ['x', 'y', 'z']),
        **ScopeTools.pack_unpack(1, ['x', 'y', 'z']),
    }), check_args=[()])

def test_cps_var_del():
    check_cps_transform('''
    def fn(flag):
        x = 3
        if flag:
            del x
        sum([])
        return x
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def fn(flag, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        x = 3
        def _cont_0(_scope_0):
            {Placeholder.new('unpack_0')}
            {Placeholder.new('pack_1')}
            def _cont_1(_res_1):
                {Placeholder.new('unpack_1')}
                _res_1
                return lambda: _cont(x)
            return lambda: _cps.interpret(sum, cont=_cont_1, stack=_stack, func_src=__func_src, locals_=_locals_1, call_id=1, lineno=4)([])
        if flag:
            del x
            {Placeholder.new('pack_0')}
            return lambda: _cont_0(_scope_0)
        else:
            {Placeholder.new('pack_0')}
            return lambda: _cont_0(_scope_0)

    ''', {
        **ScopeTools.pack_unpack(0, ['flag', 'x']),
        **ScopeTools.pack_unpack(1, ['flag', 'x']),
    }), check_args=[()])

def test_cps_desugared_calls():
    check_cps_transform('''
    def fn():
        __v0 = sum([2, 3])
        __v1 = sum([1, __v0])
        return __v1
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, 'def fn():\\n    __v0 = sum([2, 3])\\n    __v1 = sum([1, __v0])\\n    return __v1')
    def fn(*, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = 'def fn():\\n    __v0 = sum([2, 3])\\n    __v1 = sum([1, __v0])\\n    return __v1'
        {Placeholder.new('pack_0')}

        def _cont_0(_res_0):
            {Placeholder.new('unpack_0')}
            __v0 = _res_0
            {Placeholder.new('pack_1')}

            def _cont_1(_res_1):
                {Placeholder.new('unpack_1')}
                __v1 = _res_1
                return lambda : _cont(__v1)
            return lambda : _cps.interpret(sum, cont=_cont_1, stack=_stack, func_src=__func_src, locals_=_locals_1, call_id=1, lineno=2)([1, __v0])
        return lambda : _cps.interpret(sum, cont=_cont_0, stack=_stack, func_src=__func_src, locals_=_locals_0, call_id=0, lineno=1)([2, 3])
    ''', {
        **ScopeTools.pack_unpack(0, ['__v0', '__v1']),
        **ScopeTools.pack_unpack(1, ['__v0', '__v1']),
    }), check_args=[()])

def test_cps_if():
    check_cps_transform('''
    def fn(x):
        if x == 0:
            y = 0
        else:
            y = 1
        return y * 2
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, 'def fn(x):\\n    if x == 0:\\n        y = 0\\n    else:\\n        y = 1\\n    return y * 2')
    def fn(x, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = 'def fn(x):\\n    if x == 0:\\n        y = 0\\n    else:\\n        y = 1\\n    return y * 2'
        def _cont_0(_scope_0):
            {Placeholder.new('unpack_0')}
            return lambda : _cont(y * 2)
        if x == 0:
            y = 0
            {Placeholder.new('pack_0')}
            return lambda : _cont_0(_scope_0)
        else:
            y = 1
            {Placeholder.new('pack_0')}
            return lambda : _cont_0(_scope_0)
    ''', {
        **ScopeTools.pack_unpack(0, ['x', 'y'])
    }), check_args=[(0,), (1,)])

def test_cps_if_nested_call():
    check_cps_transform('''
    def fn(x):
        if x == 0:
            y = 0
        else:
            y = sum([3, x])
        return y * 2
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, 'def fn(x):\\n    if x == 0:\\n        y = 0\\n    else:\\n        y = sum([3, x])\\n    return y * 2')
    def fn(x, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = 'def fn(x):\\n    if x == 0:\\n        y = 0\\n    else:\\n        y = sum([3, x])\\n    return y * 2'

        def _cont_1(_scope_1):
            {Placeholder.new('unpack_1')}
            return lambda : _cont(y * 2)
        if x == 0:
            y = 0
            {Placeholder.new('pack_1')}
            return lambda : _cont_1(_scope_1)
        else:
            {Placeholder.new('pack_0')}

            def _cont_0(_res_0):
                {Placeholder.new('unpack_0')}
                y = _res_0
                {Placeholder.new('pack_1')}
                return lambda : _cont_1(_scope_1)
            return lambda : _cps.interpret(sum, cont=_cont_0, stack=_stack, func_src=__func_src, locals_=_locals_0, call_id=0, lineno=4)([3, x])
    ''', {
        **ScopeTools.pack_unpack(0, ['x', 'y']),
        **ScopeTools.pack_unpack(1, ['x', 'y']),
    }), check_args=[(0,), (1,)])


def test_cps_if_call_in_node_test():
    check_cps_transform('''
    def fn(x):
        if 0 == sum(x):
            y = 0
        else:
            y = 1
        return y * 2
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def fn(x, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        {Placeholder.new('pack_0')}

        def _cont_0(_res_0):
            {Placeholder.new('unpack_0')}
            def _cont_1(_scope_1):
                {Placeholder.new('unpack_1')}
                return lambda : _cont(y * 2)
            if 0 == _res_0:
                y = 0
                {Placeholder.new('pack_1')}
                return lambda : _cont_1(_scope_1)
            else:
                y = 1
                {Placeholder.new('pack_1')}
                return lambda : _cont_1(_scope_1)
        return lambda : _cps.interpret(sum, cont=_cont_0, stack=_stack, func_src=__func_src, locals_=_locals_0, call_id=0, lineno=1)(x)
    ''', {
        **ScopeTools.pack_unpack(0, ['x', 'y']),
        **ScopeTools.pack_unpack(1, ['x', 'y']),
    }), check_args=[([-1, 1],), ([3],)])


def test_cps_while_simple():
    check_cps_transform('''
    def fib(x):
        a, b = 1, 1
        ct = 0
        while ct < x:
            a, b = b, a + b
            ct = ct + 1
        return a
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def fib(x, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        from flippy.transforms import _ValueWrapper
        (a, b) = (1, 1)
        ct = 0
        @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
        def _loop_fn_0(_scope_0, *, _stack=(), _cps=_cps, _cont=lambda val: val):
            __func_src = "$IGNORE_STRING"
            {Placeholder.new('unpack_0')}
            if ct < x:
                (a, b) = (b, a + b)
                ct = ct + 1
                {Placeholder.new('pack_0')}
                return lambda : _cps.interpret(_loop_fn_0, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
            else:
                {Placeholder.new('pack_0')}
                return lambda : _cont(_ValueWrapper(_scope_0))

        def _loop_exit_fn_0(_scope_0):
            if not isinstance(_scope_0, _ValueWrapper):
                return lambda: _cont(_scope_0)
            _scope_0 = _scope_0.value
            {Placeholder.new('unpack_0')}
            return lambda : _cont(a)
        {Placeholder.new('pack_0')}
        return lambda : _cps.interpret(_loop_fn_0, cont=_loop_exit_fn_0, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
    ''', {
        **ScopeTools.pack_unpack(0, ['a', 'b', 'ct', 'x'], hashabledict=True)
    }), check_args=[(i,) for i in range(10)])


def test_cps_while_desugared():
    check_cps_transform('''
    def fib(x):
        a, b = 1, 1
        ct = 0
        while ct < x:
            a, b = b, a + b
            ct = ct + 1
        return a
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def fib(x, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        from flippy.transforms import _ValueWrapper
        (a, b) = (1, 1)
        ct = 0
        @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
        def _loop_fn_0(_scope_0, *, _stack=(), _cps=_cps, _cont=lambda val: val):
            __func_src = "$IGNORE_STRING"
            {Placeholder.new('unpack_0')}
            if True:
                def _cont_1(_scope_1):
                    {Placeholder.new('unpack_1')}
                    (a, b) = (b, a + b)
                    ct = ct + 1
                    {Placeholder.new('pack_0')}
                    return lambda : _cps.interpret(_loop_fn_0, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
                if not ct < x:
                    {Placeholder.new('pack_0')}
                    return lambda: _cont(_ValueWrapper(_scope_0))
                    {Placeholder.new('pack_1')}
                    return lambda : _cont_1(_scope_1)
                else:
                    {Placeholder.new('pack_1')}
                    return lambda : _cont_1(_scope_1)
            else:
                {Placeholder.new('pack_0')}
                return lambda : _cont(_ValueWrapper(_scope_0))

        def _loop_exit_fn_0(_scope_0):
            if not isinstance(_scope_0, _ValueWrapper):
                return lambda: _cont(_scope_0)
            _scope_0 = _scope_0.value
            {Placeholder.new('unpack_0')}
            return lambda : _cont(a)
        {Placeholder.new('pack_0')}
        return lambda : _cps.interpret(_loop_fn_0, cont=_loop_exit_fn_0, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
    ''', {
        **ScopeTools.pack_unpack(0, ['a', 'b', 'ct', 'x'], hashabledict=True),
        **ScopeTools.pack_unpack(1, ['a', 'b', 'ct', 'x']),
    }), check_args=[(i,) for i in range(10)], desugar=True)


def test_cps_while_continue_break():
    check_cps_transform('''
    def fn():
        rv = []
        ct = 0
        while ct < 10:
            ct = ct + 1
            if ct == 7:
                break
            if ct == 3:
                continue
            rv = rv + [ct]
        return rv
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def fn(*, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        from flippy.transforms import _ValueWrapper
        rv = []
        ct = 0

        @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
        def _loop_fn_0(_scope_0, *, _stack=(), _cps=_cps, _cont=lambda val: val):
            __func_src = "$IGNORE_STRING"
            {Placeholder.new('unpack_0')}
            if ct < 10:
                ct = ct + 1

                def _cont_1(_scope_1):
                    {Placeholder.new('unpack_1')}

                    def _cont_2(_scope_2):
                        {Placeholder.new('unpack_2')}
                        rv = rv + [ct]
                        {Placeholder.new('pack_0')}
                        return lambda : _cps.interpret(_loop_fn_0, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
                    if ct == 3:
                        {Placeholder.new('pack_0')}
                        return lambda : _cps.interpret(_loop_fn_0, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
                        {Placeholder.new('pack_2')}
                        return lambda : _cont_2(_scope_2)
                    else:
                        {Placeholder.new('pack_2')}
                        return lambda : _cont_2(_scope_2)
                if ct == 7:
                    {Placeholder.new('pack_0')}
                    return lambda : _cont(_ValueWrapper(_scope_0))
                    {Placeholder.new('pack_1')}
                    return lambda : _cont_1(_scope_1)
                else:
                    {Placeholder.new('pack_1')}
                    return lambda : _cont_1(_scope_1)
            else:
                {Placeholder.new('pack_0')}
                return lambda : _cont(_ValueWrapper(_scope_0))

        def _loop_exit_fn_0(_scope_0):
            if not isinstance(_scope_0, _ValueWrapper):
                return lambda: _cont(_scope_0)
            _scope_0 = _scope_0.value
            {Placeholder.new('unpack_0')}
            return lambda : _cont(rv)
        {Placeholder.new('pack_0')}
        return lambda : _cps.interpret(_loop_fn_0, cont=_loop_exit_fn_0, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
    ''', {
        **ScopeTools.pack_unpack(0, ['ct', 'rv'], hashabledict=True),
        **ScopeTools.pack_unpack(1, ['ct', 'rv']),
        **ScopeTools.pack_unpack(2, ['ct', 'rv']),
    }), check_args=[()], check_out=[[1, 2, 4, 5, 6]])


def test_cps_while_return():
    check_cps_transform('''
    def search(arr, item):
        idx = 0
        while idx < len(arr):
            if arr[idx] == item:
                return idx
            idx += 1
        return -1
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def search(arr, item, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        from flippy.transforms import _ValueWrapper
        idx = 0
        @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
        def _loop_fn_0(_scope_0, *, _stack=(), _cps=_cps, _cont=lambda val: val):
            __func_src = "$IGNORE_STRING"
            {Placeholder.new('unpack_0')}
            if idx < len(arr):
                def _cont_1(_scope_1):
                    {Placeholder.new('unpack_1')}
                    idx += 1
                    {Placeholder.new('pack_0')}
                    return lambda: _cps.interpret(_loop_fn_0, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=2)(_scope_0)
                if arr[idx] == item:
                    return lambda: _cont(idx)
                    {Placeholder.new('pack_1')}
                    return lambda: _cont_1(_scope_1)
                else:
                    {Placeholder.new('pack_1')}
                    return lambda: _cont_1(_scope_1)
            else:
                {Placeholder.new('pack_0')}
                return lambda : _cont(_ValueWrapper(_scope_0))

        def _loop_exit_fn_0(_scope_0):
            if not isinstance(_scope_0, _ValueWrapper):
                return lambda: _cont(_scope_0)
            _scope_0 = _scope_0.value
            {Placeholder.new('unpack_0')}
            return lambda : _cont(-1)
        {Placeholder.new('pack_0')}
        return lambda : _cps.interpret(_loop_fn_0, cont=_loop_exit_fn_0, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=2)(_scope_0)
    ''', {
        **ScopeTools.pack_unpack(0, ['arr', 'idx', 'item'], hashabledict=True),
        **ScopeTools.pack_unpack(1, ['arr', 'idx', 'item']),
    }), check_args=[([1, 3, 2], i) for i in range(3)], check_out=[-1, 0, 2])


def test_cps_while_nested():
    names = ['lim', 'rv', 'x', 'y']
    exp = ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def fn(lim, *, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        from flippy.transforms import _ValueWrapper
        rv = []
        x = 0

        @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
        def _loop_fn_0(_scope_0, *, _stack=(), _cps=_cps, _cont=lambda val: val):
            __func_src = "$IGNORE_STRING"
            {Placeholder.new('unpack_0')}
            if x < lim:
                y = 0

                @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
                def _loop_fn_1(_scope_1, *, _stack=(), _cps=_cps, _cont=lambda val: val):
                    __func_src = "$IGNORE_STRING"
                    {Placeholder.new('unpack_1')}
                    if y < x:
                        rv += [(x, y)]
                        y = y + 1
                        {Placeholder.new('pack_1')}
                        return lambda : _cps.interpret(_loop_fn_1, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_1, lineno=5)(_scope_1)
                    else:
                        {Placeholder.new('pack_1')}
                        return lambda : _cont(_ValueWrapper(_scope_1))

                def _loop_exit_fn_1(_scope_1):
                    if not isinstance(_scope_1, _ValueWrapper):
                        return lambda: _cont(_scope_1)
                    _scope_1 = _scope_1.value
                    {Placeholder.new('unpack_1')}
                    x = x + 1
                    {Placeholder.new('pack_0')}
                    return lambda : _cps.interpret(_loop_fn_0, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
                {Placeholder.new('pack_1')}
                return lambda : _cps.interpret(_loop_fn_1, cont=_loop_exit_fn_1, stack=_stack, func_src=__func_src, locals_=_locals_1, lineno=5)(_scope_1)
            else:
                {Placeholder.new('pack_0')}
                return lambda : _cont(_ValueWrapper(_scope_0))

        def _loop_exit_fn_0(_scope_0):
            if not isinstance(_scope_0, _ValueWrapper):
                return lambda: _cont(_scope_0)
            _scope_0 = _scope_0.value
            {Placeholder.new('unpack_0')}
            return lambda : _cont((rv, y))
        {Placeholder.new('pack_0')}
        return lambda : _cps.interpret(_loop_fn_0, cont=_loop_exit_fn_0, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
    ''', dict(
        **ScopeTools.pack_unpack(0, names, hashabledict=True),
        **ScopeTools.pack_unpack(1, names, hashabledict=True),
    ))
    check_cps_transform('''
    def fn(lim):
        rv = []
        x = 0
        while x < lim:
            y = 0
            while y < x:
                rv += [(x, y)]
                y = y + 1
            x = x + 1
        return (rv, y)
    ''', exp, check_args=[(3,)], check_out=[
        ([(1, 0), (2, 0), (2, 1)], 2),
    ])


def test_cps_while_call_in_test():
    # This test case is checking whether a call in the while loop's test is dynamically computed.
    # In the past, we could compile that call out when desugaring, which would mean any function
    # calls would be extracted into assignments to temporary vars. Our approach was to desugar
    # while loops to have a trivial test, which works well with our desugaring scheme.
    check_cps_transform('''
    def fn():
        ct = 0
        # To avoid control flow from `and`, we use &
        while (sum([ct]) < 3) & (ct < 100):
            ct += 1
        return ct
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def fn(*, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        from flippy.transforms import _ValueWrapper
        ct = 0

        @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
        def _loop_fn_0(_scope_0, *, _stack=(), _cps=_cps, _cont=lambda val: val):
            __func_src = "$IGNORE_STRING"
            {Placeholder.new('unpack_0')}
            if True:
                {Placeholder.new('pack_1')}

                def _cont_1(_res_1):
                    {Placeholder.new('unpack_1')}
                    def _cont_2(_scope_2):
                        {Placeholder.new('unpack_2')}
                        ct = ct + 1
                        {Placeholder.new('pack_0')}
                        return lambda : _cps.interpret(_loop_fn_0, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=2)(_scope_0)
                    if not (_res_1 < 3) & (ct < 100):
                        {Placeholder.new('pack_0')}
                        return lambda: _cont(_ValueWrapper(_scope_0))
                        {Placeholder.new('pack_2')}
                        return lambda : _cont_2(_scope_2)
                    else:
                        {Placeholder.new('pack_2')}
                        return lambda : _cont_2(_scope_2)
                return lambda : _cps.interpret(sum, cont=_cont_1, stack=_stack, func_src=__func_src, locals_=_locals_1, call_id=1, lineno=3)([ct])
            else:
                {Placeholder.new('pack_0')}
                return lambda : _cont(_ValueWrapper(_scope_0))

        def _loop_exit_fn_0(_scope_0):
            if not isinstance(_scope_0, _ValueWrapper):
                return lambda: _cont(_scope_0)
            _scope_0 = _scope_0.value
            {Placeholder.new('unpack_0')}
            return lambda : _cont(ct)
        {Placeholder.new('pack_0')}
        return lambda : _cps.interpret(_loop_fn_0, cont=_loop_exit_fn_0, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=2)(_scope_0)
    ''', {
        **ScopeTools.pack_unpack(0, ['ct'], hashabledict=True),
        **ScopeTools.pack_unpack(1, ['ct']),
        **ScopeTools.pack_unpack(2, ['ct']),
    }), check_args=[()], desugar=True)


def test_cps_while_nested_call():
    check_cps_transform('''
    def fn():
        x = 0
        ct = 0
        while ct < 10:
            x = sum([x, 1])
            x = sum([x, 2])
            ct = ct + 1
        return x
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, 'def fn():\\n    x = 0\\n    ct = 0\\n    while ct < 10:\\n        x = sum([x, 1])\\n        x = sum([x, 2])\\n        ct = ct + 1\\n    return x')
    def fn(*, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = 'def fn():\\n    x = 0\\n    ct = 0\\n    while ct < 10:\\n        x = sum([x, 1])\\n        x = sum([x, 2])\\n        ct = ct + 1\\n    return x'
        from flippy.transforms import _ValueWrapper
        x = 0
        ct = 0

        @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
        def _loop_fn_0(_scope_0, *, _stack=(), _cps=_cps, _cont=lambda val: val):
            __func_src = "$IGNORE_STRING"
            {Placeholder.new('unpack_0')}
            if ct < 10:
                {Placeholder.new('pack_1')}
                def _cont_1(_res_1):
                    {Placeholder.new('unpack_1')}
                    x = _res_1
                    {Placeholder.new('pack_2')}
                    def _cont_2(_res_2):
                        {Placeholder.new('unpack_2')}
                        x = _res_2
                        ct = ct + 1
                        {Placeholder.new('pack_0')}
                        return lambda : _cps.interpret(_loop_fn_0, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
                    return lambda : _cps.interpret(sum, cont=_cont_2, stack=_stack, func_src=__func_src, locals_=_locals_2, call_id=2, lineno=5)([x, 2])
                return lambda : _cps.interpret(sum, cont=_cont_1, stack=_stack, func_src=__func_src, locals_=_locals_1, call_id=1, lineno=4)([x, 1])
            else:
                {Placeholder.new('pack_0')}
                return lambda : _cont(_ValueWrapper(_scope_0))

        def _loop_exit_fn_0(_scope_0):
            if not isinstance(_scope_0, _ValueWrapper):
                return lambda: _cont(_scope_0)
            _scope_0 = _scope_0.value
            {Placeholder.new('unpack_0')}
            return lambda : _cont(x)
        {Placeholder.new('pack_0')}
        return lambda : _cps.interpret(_loop_fn_0, cont=_loop_exit_fn_0, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
    ''', dict(
        **ScopeTools.pack_unpack(0, ['ct', 'x'], hashabledict=True),
        **ScopeTools.pack_unpack(1, ['ct', 'x']),
        **ScopeTools.pack_unpack(2, ['ct', 'x']),
    )), check_args=[()])


def test_cps_for():
    check_cps_transform('''
    def fn():
        rv = []
        for x in list(range(10)):
            rv += [x**2]
        return rv
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def fn(*, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        from flippy.transforms import _ValueWrapper
        rv = []
        {Placeholder.new('pack_1')}
        def _cont_1(_res_1):
            {Placeholder.new('unpack_1')}
            {Placeholder.new('pack_0')}
            def _cont_0(_res_0):
                {Placeholder.new('unpack_0')}
                _for_iter_4 = _res_0
                _for_idx_4 = 0

                @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
                def _loop_fn_2(_scope_2, *, _stack=(), _cps=_cps, _cont=lambda val: val):
                    __func_src = "$IGNORE_STRING"
                    {Placeholder.new('unpack_2')}
                    if True:
                        {Placeholder.new('pack_3')}
                        def _cont_3(_res_3):
                            {Placeholder.new('unpack_3')}

                            def _cont_4(_scope_4):
                                {Placeholder.new('unpack_4')}
                                x = _for_iter_4[_for_idx_4]
                                _for_idx_4 = _for_idx_4 + 1
                                rv = rv + [x ** 2]
                                {Placeholder.new('pack_2')}
                                return lambda : _cps.interpret(_loop_fn_2, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_2, lineno=4)(_scope_2)
                            if not _for_idx_4 < _res_3:
                                {Placeholder.new('pack_2')}
                                return lambda: _cont(_ValueWrapper(_scope_2))
                                {Placeholder.new('pack_4')}
                                return lambda : _cont_4(_scope_4)
                            else:
                                {Placeholder.new('pack_4')}
                                return lambda : _cont_4(_scope_4)
                        return lambda : _cps.interpret(len, cont=_cont_3, stack=_stack, func_src=__func_src, locals_=_locals_3, call_id=3, lineno=5)(_for_iter_4)
                    else:
                        {Placeholder.new('pack_2')}
                        return lambda: _cont(_ValueWrapper(_scope_2))

                def _loop_exit_fn_2(_scope_2):
                    if not isinstance(_scope_2, _ValueWrapper):
                        return lambda: _cont(_scope_2)
                    _scope_2 = _scope_2.value
                    {Placeholder.new('unpack_2')}
                    return lambda : _cont(rv)
                {Placeholder.new('pack_2')}
                return lambda : _cps.interpret(_loop_fn_2, cont=_loop_exit_fn_2, stack=_stack, func_src=__func_src, locals_=_locals_2, lineno=4)(_scope_2)
            return lambda : _cps.interpret(list, cont=_cont_0, stack=_stack, func_src=__func_src, locals_=_locals_0, call_id=0, lineno=2)(_res_1)
        return lambda : _cps.interpret(range, cont=_cont_1, stack=_stack, func_src=__func_src, locals_=_locals_1, call_id=1, lineno=2)(10)
    ''', {
        **ScopeTools.pack_unpack(0, ['_for_idx_4', '_for_iter_4', 'rv', 'x']),
        **ScopeTools.pack_unpack(1, ['_for_idx_4', '_for_iter_4', 'rv', 'x']),
        **ScopeTools.pack_unpack(2, ['_for_idx_4', '_for_iter_4', 'rv', 'x'], hashabledict=True),
        **ScopeTools.pack_unpack(3, ['_for_idx_4', '_for_iter_4', 'rv', 'x']),
        **ScopeTools.pack_unpack(4, ['_for_idx_4', '_for_iter_4', 'rv', 'x']),
    }), check_args=[()], desugar=True)


def test_cps_for_target_in_scope():
    check_cps_transform(f'''
    def fn():
        for x in [1, 2, 3]:
            if x == 3:
                break
        return x
    ''', '', check_args=[()], desugar=True, compare_ast=False)

    check_cps_transform(f'''
    def fn():
        for x in [1, 2, 3]:
            if x == 3:
                return x
    ''', '', check_args=[()], desugar=True, compare_ast=False)

    # loop where variable is defined inside loop
    check_cps_transform(f'''
    def fn(val):
        idx = 0
        while True:
            if idx == 0:
                x = 0
            else: # idx > 0
                x += val
            idx += 1
            if idx >= 3:
                break
        return x
    ''', '', check_args=[(4,)], desugar=True, compare_ast=False)

    # issue returning variable declared in loop
    for var in ['outer', 'inner']:
        check_cps_transform(f'''
        def fn():
            outer = 0
            for inner in [1, 2, 3]:
                outer = inner
            return {var}
        ''', '', check_args=[()], desugar=True, compare_ast=False)

    # similar to above, but also checking output
    check_cps_transform('''
    def fn():
        for x in [1, 2, 3]:
            pass
        return x
    ''', ScopeTools.fill(f'''
    @lambda fn: CPSFunction(fn, '$FN_SOURCE')
    def fn(*, _stack=(), _cps=_cps, _cont=lambda val: val):
        __func_src = '$FN_SOURCE'
        from flippy.transforms import _ValueWrapper
        _for_iter_3 = [1, 2, 3]
        _for_idx_3 = 0
        @lambda fn: CPSFunction(fn, "$IGNORE_STRING")
        def _loop_fn_0(_scope_0, *, _stack=(), _cps=_cps, _cont=lambda val: val):
            __func_src = "$IGNORE_STRING"
            {Placeholder.new('unpack_0')}
            if True:
                {Placeholder.new('pack_1')}
                def _cont_1(_res_1):
                    {Placeholder.new('unpack_1')}
                    def _cont_2(_scope_2):
                        {Placeholder.new('unpack_2')}
                        x = _for_iter_3[_for_idx_3]
                        _for_idx_3 = _for_idx_3 + 1
                        pass
                        {Placeholder.new('pack_0')}
                        return lambda: _cps.interpret(_loop_fn_0, cont=_cont, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
                    if not _for_idx_3 < _res_1:
                        {Placeholder.new('pack_0')}
                        return lambda: _cont(_ValueWrapper(_scope_0))
                        {Placeholder.new('pack_2')}
                        return lambda: _cont_2(_scope_2)
                    else:
                        {Placeholder.new('pack_2')}
                        return lambda: _cont_2(_scope_2)
                return lambda: _cps.interpret(len, cont=_cont_1, stack=_stack, func_src=__func_src, locals_=_locals_1, call_id=1, lineno=4)(_for_iter_3)
            else:
                {Placeholder.new('pack_0')}
                return lambda: _cont(_ValueWrapper(_scope_0))

        def _loop_exit_fn_0(_scope_0):
            if not isinstance(_scope_0, _ValueWrapper):
                return lambda: _cont(_scope_0)
            _scope_0 = _scope_0.value
            {Placeholder.new('unpack_0')}
            return lambda: _cont(x)
        {Placeholder.new('pack_0')}
        return lambda: _cps.interpret(_loop_fn_0, cont=_loop_exit_fn_0, stack=_stack, func_src=__func_src, locals_=_locals_0, lineno=3)(_scope_0)
    ''', {
        **ScopeTools.pack_unpack(0, ['_for_idx_3', '_for_iter_3', 'x'], hashabledict=True),
        **ScopeTools.pack_unpack(1, ['_for_idx_3', '_for_iter_3', 'x']),
        **ScopeTools.pack_unpack(2, ['_for_idx_3', '_for_iter_3', 'x']),
        **ScopeTools.pack_unpack(3, ['_for_idx_3', '_for_iter_3', 'x']),
    }), check_args=[()], desugar=True)


def _code_diff(a, b):
    from difflib import ndiff
    def _process_string(s):
        return (s.strip() + '\n').splitlines(keepends=True) # Make sure we have a newline at the very end
    diff = ndiff(_process_string(a), _process_string(b))
    return ''.join(diff)


def check_cps_transform(src, exp_src, *, check_args=[], check_out=None, desugar=False, compare_ast=True):
    src = textwrap.dedent(src)
    node = ast.parse(src)
    if desugar:
        node = DesugaringTransform()(node)
    # Replace $FN_SOURCE before we transform
    exp_src = exp_src.replace('$FN_SOURCE', ast.unparse(node).replace('\n', '\\n'))
    node = CPSTransform()(node)

    exp_src = textwrap.dedent(exp_src)

    if compare_ast:
        exp_node = ast.parse(exp_src)
        assert_compare_ast(node, exp_node)
    else:
        # Need to do this since we don't have transformed source.
        exp_node = node

    assert (
        isinstance(node, ast.Module) and
        len(node.body) == 1 and
        isinstance(node.body[0], ast.FunctionDef)
    )
    fn_name = node.body[0].name

    src_context = {'CPSFunction': CPSFunction}
    exec(src, src_context)

    exp_context = {'CPSFunction': CPSFunction}
    interpreter = CPSInterpreter()
    exp_context[CPSTransform.cps_interpreter_name] = interpreter
    exec(interpreter.compile('test.py', exp_node), exp_context)

    assert check_args, 'Must supply test case to test transformed function.'
    if check_out is not None:
        assert len(check_out) == len(check_args)
    for idx, args in enumerate(check_args):
        try:
            expected = src_context[fn_name](*args)
        except Exception as err:
            expected = err
        if check_out is not None:
            assert expected == check_out[idx]
        # We execute using only a simple trampoline, so this implementation can't handle stochastic primitives.
        try:
            actual = trampoline(exp_context[fn_name](*args))
        except Exception as err:
            actual = err

        if isinstance(expected, Exception):
            assert type(expected) == type(actual) and str(expected) == str(actual)
        else:
            assert expected == actual

def assert_compare_ast(node, exp_node):
    '''
    Helper function with nice error message for debugging.
    '''
    assert compare_ast(node, exp_node), f'Diff from expected source to transformed source.\n{_code_diff(ast.unparse(exp_node), ast.unparse(node))}\n\nTransformed source:\n{ast.unparse(node)}'

def compare_ast(node1, node2):
    if type(node1) is not type(node2):
        return False
    if isinstance(node1, ast.AST):
        for k, v in vars(node1).items():
            if k in ('lineno', 'col_offset', 'end_col_offset', 'end_lineno'):
                continue
            if not compare_ast(v, getattr(node2, k)):
                return False
        return True
    elif isinstance(node1, list):
        return all(itertools.starmap(compare_ast, itertools.zip_longest(node1, node2)))
    else:
        # A special case to avoid including function sources.
        if isinstance(node1, str) and '$IGNORE_STRING' in (node1, node2):
            return True
        return node1 == node2

def test_Placeholder():
    trans = Placeholder.fill(ast.parse(textwrap.dedent(f'''
        x = {Placeholder.new('expr')}
        {Placeholder.new('stmt')}
        {Placeholder.new('stmts')}
    ''')), dict(
        expr=ast.Constant(3),
        stmt=ast.parse('y = f()').body[0],
        stmts=ast.parse('z=4\na=5').body,
    ))
    expected = ast.parse(textwrap.dedent('''
        x = 3
        y = f()
        z = 4
        a = 5
    '''))
    assert_compare_ast(trans, expected)



from typing import Callable, Generic, TypeVar, Dict, Any, Tuple, List, Union, Optional
from abc import ABC, abstractmethod
from collections import defaultdict
import math

from flippy.distributions import Distribution, Element
from flippy.distributions.builtin_dists import Categorical
from flippy.tools import PackagePlaceholder

try:
    import numpy as np
except ImportError:
    np = PackagePlaceholder("numpy")

MarginalLikelihood = float

class InferenceAlgorithm(ABC, Generic[Element]):
    @abstractmethod
    def run(self, *args, **kws) -> Distribution[Element]:
        raise NotImplementedError

    @property
    def is_cachable(self) -> bool:
        raise NotImplementedError

class InferenceResult(ABC, Distribution[Element]):
    @property
    @abstractmethod
    def marginal_likelihood(self) -> float:
        raise NotImplementedError

class DiscreteInferenceResult(InferenceResult, Categorical[Element]):
    def __init__(self, support, probabilities, marginal_likelihood):
        Categorical.__init__(self, support=support, probabilities=probabilities)
        self._marginal_likelihood = marginal_likelihood

    @classmethod
    def from_values_scores(
        cls,
        return_values: List[Element],
        return_scores: List[float]
    ) -> 'DiscreteInferenceResult':
        assert len(return_values) == len(return_scores)
        try:
            return cls._from_values_scores_numpy(return_values, return_scores)
        except ImportError:
            return cls._from_values_scores_builtin(return_values, return_scores)

    @classmethod
    def _from_values_scores_numpy(
        cls,
        return_values: List[Element],
        return_scores: List[float]
    ) -> 'DiscreteInferenceResult':
        return_scores = np.array(return_scores)
        max_score = np.max(return_scores)
        return_probs = np.exp(return_scores - max_score)
        return_probs = return_probs / np.sum(return_probs)
        values_probs = defaultdict(float)
        for value, prob in zip(return_values, return_probs):
            values_probs[value] += prob
        values, probs = zip(*values_probs.items())
        log_marginal_likelihood = max_score + np.log(np.sum(np.exp(return_scores - max_score)))
        marginal_likelihood = np.exp(log_marginal_likelihood)
        return cls(
            support=values,
            probabilities=probs,
            marginal_likelihood=marginal_likelihood
        )

    @classmethod
    def _from_values_scores_builtin(
        cls,
        return_values: List[Element],
        return_scores: List[float]
    ) -> 'DiscreteInferenceResult':
        max_score = max(return_scores)
        return_probs = [math.exp(score - max_score) for score in return_scores]
        total_prob = sum(return_probs)
        return_probs = [prob / total_prob for prob in return_probs]
        values_probs = defaultdict(float)
        for value, prob in zip(return_values, return_probs):
            values_probs[value] += prob
        values, probs = zip(*values_probs.items())
        log_marginal_likelihood = max_score + math.log(total_prob)
        marginal_likelihood = math.exp(log_marginal_likelihood)
        return cls(
            support=values,
            probabilities=probs,
            marginal_likelihood=marginal_likelihood
        )

    @property
    def marginal_likelihood(self) -> float:
        return self._marginal_likelihood



"""
Distributions are used to represent random variables in probabilistic models.
FlipPy provides several built-in distributions, and new distributions can be
defined by subclassing the `Distribution` base class.
"""

from flippy.distributions.base import Distribution, Element
from flippy.distributions.random import RandomNumberGenerator, default_rng
from flippy.distributions.builtin_dists import Bernoulli, Categorical, Multinomial, \
    Gaussian, Normal, Gamma, Uniform, Beta, Binomial, Geometric, Poisson, \
    BetaBinomial, Dirichlet, DirichletMultinomial, Mixture
from flippy.distributions.scipy_dists import MultivariateNormal, InverseWishart, \
    NormalNormal, MultivariateNormalNormal

__all__ = [
    "RandomNumberGenerator",
    "Distribution",

    "Bernoulli",
    "Categorical",
    "Multinomial",
    "Gaussian",
    "Normal",
    "Gamma",
    "Uniform",
    "Beta",
    "Binomial",
    "Geometric",
    "Poisson",
    "BetaBinomial",
    "Dirichlet",
    "DirichletMultinomial",
    "NormalNormal",
    "MultivariateNormal",
    "InverseWishart",
    "MultivariateNormalNormal",

    "Mixture"
]

class ZeroDistributionError(Exception):
    pass




import pytest
from flippy.hashable import hashabledict, hashablelist, hashableset

def test_hashableset():
    hs = hashableset([1,2,3])
    s = set([1,2,3])
    assert hs == s, "Equality depends only on contents"
    hs2 = hashableset(s)
    assert hash(hs) == hash(hs2), "Hash equality depends only on contents"
    assert type(hs | {4}) == hashableset, "Adding a set to a hashableset returns a hashableset"
    assert type({4} | hs ) == hashableset, "Adding a set to a hashableset returns a hashableset"
    assert type(hs2 & {4}) == hashableset, "Intersecting a hashableset with a set returns a hashableset"
    assert type({4} & hs2) == hashableset, "Intersecting a hashableset with a set returns a hashableset"
    assert {hs: 1} == {hs2: 1}, "Hashableset is hashable"
    assert {hs: 1} != {hashableset(hs2 | {4}): 2}, "Hashableset is hashable"

    with pytest.raises(TypeError):
        hs.update([4])
    with pytest.raises(TypeError):
        hs.intersection_update([4])
    with pytest.raises(TypeError):
        hs |= [4]
    with pytest.raises(TypeError):
        hs &= [4]
    with pytest.raises(TypeError):
        hs ^= [4]
    with pytest.raises(TypeError):
        hs.difference_update([4])
    with pytest.raises(TypeError):
        hs.symmetric_difference_update([4])
    with pytest.raises(TypeError):
        hs.add(4)
    with pytest.raises(TypeError):
        hs.remove(4)
    with pytest.raises(TypeError):
        hs.discard(4)
    with pytest.raises(TypeError):
        hs.pop()
    with pytest.raises(TypeError):
        hs.clear()

def test_hashablelist():
    hl = hashablelist([1,2,3])
    l = [1,2,3]
    assert hl == l, "Equality depends only on contents"
    hl2 = hashablelist(l)
    assert hash(hl) == hash(hl2), "Hash equality depends only on contents"
    assert type(hl + [4]) == hashablelist, "Adding a list to a hashablelist returns a hashablelist"
    assert type([4] + hl) == hashablelist, "Adding a list to a hashablelist returns a hashablelist"
    assert type(hl2 * 3) == hashablelist, "Multiplying a hashablelist returns a list"
    assert type(3*hl2) == hashablelist, "Multiplying a hashablelist returns a list"
    assert {hl: 1} == {hl2: 1}, "Hashablelist is hashable"
    assert {hl: 1} != {hashablelist(hl2 + [4]): 2}, "Hashablelist is hashable"

    # Test immutability guardrails
    with pytest.raises(TypeError):
        hl[0] = 3
    with pytest.raises(TypeError):
        del hl[0]
    with pytest.raises(TypeError):
        hl.append(4)
    with pytest.raises(TypeError):
        hl.insert(0, 4)
    with pytest.raises(TypeError):
        hl.pop()
    with pytest.raises(TypeError):
        hl.remove()
    with pytest.raises(TypeError):
        hl.clear()
    with pytest.raises(TypeError):
        hl += [4]
    with pytest.raises(TypeError):
        hl.sort()
    with pytest.raises(TypeError):
        hl *= 2
    with pytest.raises(TypeError):
        hl.reverse()

def test_hashabledict():
    hd = hashabledict({'a': 1, 'b': 2})
    d = {'b': 2, 'a': 1}
    assert hd == d, "Equality depends only on keys and values, not order or hashability"
    hd2 = hashabledict(d)
    assert hash(hd) == hash(hd2), "Hashes are equal for equal keys and values"
    assert type(hd | {'c': 3}) == hashabledict, "Union of hashabledict is hashabledict"
    assert type({'c': 3} | hd) == hashabledict, "Union of hashabledict is hashabledict"
    assert {hd: 1} == {hd: 1}, "hashabledict is hashable"
    assert {hd: 1} != {hashabledict(hd | {'c': 3}): 1}

    # Test immutability guardrails
    with pytest.raises(TypeError):
        hd['c'] = 3
    with pytest.raises(TypeError):
        del hd['a']
    with pytest.raises(TypeError):
        hd.update({'c': 3})
    with pytest.raises(TypeError):
        hd.pop('a')
    with pytest.raises(TypeError):
        hd.popitem()
    with pytest.raises(TypeError):
        hd.clear()
    with pytest.raises(TypeError):
        hd.setdefault('c', 3)
    with pytest.raises(TypeError):
        hd.fromkeys('abc') # this uses .__setitem__
    with pytest.raises(TypeError):
        hd |= {'c': 3}

def test_recursive_coercing():
    hash(hashabledict({'a': (), 'k': {}}))
    hash(hashabledict({'a': [1,2,3], 'b': {'c': {4,5,6}}, 'd': {'e': 7}}))
    hash(hashablelist([1,2,3, [4,5,6], {7,8,9}, {'a': 10, 'b': {}}]))



from typing import Any, Callable, Dict, Tuple, Union, TYPE_CHECKING, Protocol, Hashable, TypeVar

if TYPE_CHECKING:
    from flippy.interpreter import StackFrame
    from flippy.distributions import Distribution
    from flippy.core import ProgramState
    from flippy.distributions.random import RandomNumberGenerator, default_rng
    from flippy.interpreter import CPSInterpreter, Stack

Element = TypeVar("Element")

VariableName = Hashable
SampleValue = Any
ReturnValue = Any

# Thunks and continuations, combined with the CPSInterpreter, form the building
# blocks of the ProgramState abstraction used for defining inference algorithms.

# A thunk is a function that takes no arguments and represents some point in the
# computation that, when executed, will either step to the next point in the
# computation (return a new thunk) or return a ProgramState that can be used to
# modify or inspect the state of the computation.
# By executing a series of thunks using a trampoline, we can run a program.
Thunk = Callable[[], Union['Thunk', 'ProgramState']]

# In continuation-passing style (CPS), a continuation is a parameterizable function that represents
# "what should be done next" after the current function finishes
# and computes a value for the continuation to use
# (i.e., how the program should continue executing).
# Continuations provide a way to represent the stack of a program explicitly.

# In our implementation, continuations either return Thunks to be executed by
# a trampoline, or they return a ProgramState that can be used to modify or inspect
# the computation.
Continuation = Callable[..., Union[Thunk, 'ProgramState']]
NonCPSCallable = Callable[..., Any]

# Type annotation for a sample method of a Distribution
class SampleCallable(Protocol[Element]):
    __self__: 'Distribution[Element]'
    def __call__(self, rng: 'RandomNumberGenerator', name: Hashable) -> 'Element':
        ...

# Type annotation for a observe method of a Distribution
class ObserveCallable(Protocol[Element]):
    __self__: 'Distribution[Element]'
    def __call__(self, value: Element) -> None:
        ...

class Method(Protocol):
    __self__: Any
    __func__: Callable
    __name__: str

# Type annotation for a CPS-transformed function
class CPSCallable(Protocol):
    def __call__(
        self,
        *args,
        _cps: 'CPSInterpreter',
        _stack: 'Stack',
        _cont: 'Continuation',
        **kws
    ) -> Union['Thunk', 'ProgramState']:
        ...



'''
FlipPy lets you specify probabilistic programs in Python syntax
while seamlessly interacting with the rest of Python.

# Quick start

```bash
pip install flippy-lang
```

# Example: Sum of bernoullis

```python
from flippy import infer, flip

@infer
def fn():
    x = flip(0.5)
    y = flip(0.5)
    return x + y

fn() # Distribution({0: 0.25, 1: 0.5, 2: 0.25})
```

# Documentation

Here is the documentation for writing models in FlipPy.
- The core API for declaring a model ([link](#api))
- Specifying distributions ([link](flippy/distributions))
- Selecting inference algorithms ([link](flippy/inference))

# Tutorials

- [Introductory tutorial](tutorials/00-intro)
- [Rational Speech Acts (RSA)](tutorials/01-RSA)
- [Language of Thought (LoT)](tutorials/02-LoT)
- [Hidden Markov Models (HMMs)](tutorials/03-HMMs)
- [Bayesian Non-parametrics](tutorials/04-DP-MM)
- [Intuitive Physics](tutorials/05-Physics)
- [Sequential Decision-Making](tutorials/06-Sequential-DM)

# API

'''

import math
from typing import Callable, Sequence, Union, TypeVar, overload, Generic

from flippy.transforms import CPSTransform
from flippy.inference import \
    SimpleEnumeration, Enumeration, SamplePrior, MetropolisHastings, \
    LikelihoodWeighting, InferenceAlgorithm
from flippy.distributions import Categorical, Bernoulli, Distribution, Uniform,\
    Element, Normal
from flippy.distributions.random import default_rng
from flippy.distributions.base import _factor_dist
from flippy.core import global_store
from flippy.hashable import hashabledict
from flippy.map import recursive_map
from flippy.tools import LRUCache

from flippy.interpreter import CPSInterpreter, keep_deterministic, \
    cps_transform_safe_decorator, DescriptorMixIn

__all__ = [
    'infer',

    'flip',
    'draw_from',
    'uniform',
    'normal',

    'factor',
    'condition',
    # 'map_observe',
    'keep_deterministic',
    'mem',

    # submodules

    # Model specification
    'distributions',
    # Inference algorithms
    'inference',

    # Execution model
    'core',
    'callentryexit',
    # 'map',
]

class InferCallable(Generic[Element], DescriptorMixIn):
    '''
    @private
    '''
    def __init__(
        self,
        func: Callable[..., Element],
        method : Union[type[InferenceAlgorithm], str] = "Enumeration",
        cache_size=0,
        **kwargs
    ):
        DescriptorMixIn.__init__(self, func)

        if isinstance(method, str):
            method : type[InferenceAlgorithm] = {
                'Enumeration': Enumeration,
                'SimpleEnumeration': SimpleEnumeration,
                'SamplePrior': SamplePrior,
                'MetropolisHastings': MetropolisHastings,
                'LikelihoodWeighting' : LikelihoodWeighting
            }[method]
        self.cache_size = cache_size
        self.cache = LRUCache(cache_size)
        self.method = method
        self.kwargs = kwargs
        self.func = func
        self.inference_alg = None
        setattr(self, CPSTransform.is_transformed_property, True)

    def _lazy_init(self):
        if self.inference_alg is not None:
            return
        func = self.func
        if isinstance(func, (classmethod, staticmethod)):
            func = func.__func__
        if not CPSTransform.is_transformed(func):
            func = CPSInterpreter().non_cps_callable_to_cps_callable(func)
        self.inference_alg = self.method(func, **self.kwargs)

        if not self.inference_alg.is_cachable:
            self.cache_size = 0

    def __call__(self, *args, _cont=None, _cps=None, _stack=None, **kws) -> Distribution[Element]:
        self._lazy_init()
        if self.cache_size > 0:
            kws_tuple = tuple(sorted(kws.items()))
            if (args, kws_tuple) in self.cache:
                dist = self.cache[args, kws_tuple]
            else:
                dist = self.inference_alg.run(*args, **kws)
                self.cache[args, kws_tuple] = dist
        else:
            dist = self.inference_alg.run(*args, **kws)
        if _cont is None:
            return dist
        else:
            return lambda : _cont(dist)

def infer(
    func: Callable[..., Element]=None,
    method=Enumeration,
    cache_size=1024,
    **kwargs
) -> InferCallable[Element]:
    '''
    Turns a function into a stochastic function, that represents a posterior distribution.

    This is the main interface for performing inference in FlipPy.

    - `method` specifies the inference method and can either be an instance of
    an `InferenceAlgorithm` or a string. Defaults to `Enumeration`.
    - `**kwargs` are keyword arguments passed to the inference method.
    '''
    return InferCallable(func, method, cache_size, **kwargs)
infer = cps_transform_safe_decorator(infer)

# type hints for infer - if we can use ParamSpecs this will be cleaner
InferenceType = Callable[[Callable[..., Element]], InferCallable[Element]]
infer : Callable[..., Union[InferCallable, InferenceType]]

def recursive_filter(fn, iter):
    if not iter:
        return []
    if fn(iter[0]):
        head = [iter[0]]
    else:
        head = []
    return head + recursive_filter(fn, iter[1:])

def recursive_reduce(fn, iter, initializer):
    if len(iter) == 0:
        return initializer
    return recursive_reduce(fn, iter[1:], fn(initializer, iter[0]))

def factor(score):
    '''
    Adds a real-valued `score` (i.e., log-probability) to the weight of the
    current trace.
    '''
    _factor_dist.observe(score)

def condition(cond: float):
    '''
    Used for conditioning statements. When `cond` is a boolean, this behaves like
    typical conditioning.

    - `cond` is a non-negative multiplicative weight for the conditioning. When zero,
        the trace is assigned zero probability.
    '''
    if cond == 0:
        _factor_dist.observe(-float("inf"))
    else:
        _factor_dist.observe(math.log(cond))

def flip(p=.5, name=None):
    '''
    Samples from a Bernoulli distribution with probability `p`.
    '''
    return bool(Bernoulli(p).sample(name=name))

@keep_deterministic
def _draw_from_dist(n: Union[Sequence[Element], int]) -> Distribution[Element]:
    if isinstance(n, int):
        return Categorical(range(n))
    if hasattr(n, '__getitem__'):
        return Categorical(n)
    else:
        return Categorical(list(n))

@overload
def draw_from(n: int) -> int:
    ...
@overload
def draw_from(n: Sequence[Element]) -> Element:
    ...
def draw_from(n: Union[Sequence[Element], int]) -> Element:
    '''
    Samples uniformly from `n` when it is a sequence.
    When `n` is an integer, a sample is drawn from `range(n)`.
    '''
    return _draw_from_dist(n).sample()

def mem(fn: Callable[..., Element]) -> Callable[..., Element]:
    '''
    Turns a function into a stochastically memoized function.
    Stores information in trace-specific storage.
    '''
    def mem_wrapper(*args, **kws):
        key = (fn, args, tuple(sorted(kws.items())))
        kws = hashabledict(kws)
        if key in global_store:
            return global_store.get(key)
        else:
            value = fn(*args, **kws)
            global_store.set(key, value)
            return value
    return mem_wrapper
mem = cps_transform_safe_decorator(mem)

_uniform = Uniform()
def uniform():
    '''
    Samples from a uniform distribution over the interval $[0, 1]$.
    '''
    return _uniform.sample()

_normal = Normal(0, 1)
def normal(mean=0, std=1):
    '''
    Samples from a standard normal distribution.
    '''
    return mean + std * _normal.sample()

@keep_deterministic
def map_log_probability(distribution: Distribution[Element], values: Sequence[Element]) -> float:
    return sum(distribution.log_probability(i) for i in values)

def map_observe(distribution: Distribution[Element], values: Sequence[Element]) -> float:
    """
    Calculates the total log probability of a sequence of
    independent values from a distribution.
    """
    log_prob = map_log_probability(distribution, values)
    factor(log_prob)
    return log_prob



from flippy.transforms import PythonSubsetValidator
import ast
import textwrap
import pytest

def _validate(source):
    v = PythonSubsetValidator()
    source = textwrap.dedent(source)
    node = ast.parse(source)
    v(node, source)

def _invalid(source, exception_code):
    with pytest.raises(SyntaxError) as err:
        _validate(source)

    assert 'Found unsupported Python feature.' in str(err)

    # Now check exception content to make sure it references code.
    _, exc, tb = err._excinfo
    assert exception_code in exc.text
    # Can make this check more extensive by borrowing from test_interpreter.py

def test_validator():
    _invalid('def fn():\n  global x', 'global x')
    _invalid('def fn():\n  nonlocal x', 'nonlocal x')

    _invalid('class X(object):\n  pass', 'class X')
    _invalid('async def fn():\n  pass', 'async def fn')

    _invalid('async for x in range(3): pass', 'async for x')
    _invalid('async with x: pass', 'async with')
    _invalid('with x: pass', 'with')

    if hasattr(ast, 'Match'):
        _invalid('match None:\n\tcase _: pass', 'match')

    _invalid('''
    try:
        pass
    except:
        pass
    ''', 'try')
    if hasattr(ast, 'TryStar'):
        _invalid('''
        try:
            pass
        except* Exception as e:
            pass
        ''', 'try')

    _invalid('import x', 'import')
    _invalid('from x import x', 'import')

    _invalid('await x', 'await')
    _invalid('yield x', 'yield x')
    _invalid('yield from x', 'yield from x')
    _invalid('if x := 3: pass', ':=')

    for loop in [
        'for x in y:',
        'while True:',
    ]:
        for fn, name in [
            ('def z(): pass', 'def'),
            ('z = lambda: None', 'lambda'),
        ]:
            _invalid(f'''
            {loop}
                {fn}
            ''', name)

    # Making sure flag is not unset
    _invalid('''
    for x in y:
        for x in y:
            pass
        def z(): pass
    ''', 'def')

def test_assignment():
    _validate('x = 3')
    _validate('del x')
    _validate('x += 3')
    _validate('x: int')
    _validate('x: int = 3')

    _validate('[x, y] = range(2)')
    _validate('(x, y) = range(2)')
    _validate('[[x, y], z] = [[0, 1], 2]')

    # Testing setting from attribute to make sure ctx=Load() works.
    _validate('x = sum.__name__')

    for lhs in ['x.f', 'x[0]']:
        _invalid(f'{lhs} = 3', f'{lhs} = 3')
        _invalid(f'del {lhs}', f'del {lhs}')
        _invalid(f'{lhs} += 3', f'{lhs} += 3')
        _invalid(f'{lhs}: int = 3', f'{lhs}: int = 3')
        _invalid(f'[{lhs}] = [3]', f'[{lhs}] = [3]')



from setuptools import setup
import pathlib
curr_dir = pathlib.Path(__file__).absolute().parent

with open(curr_dir / "version.txt") as f:
    version = f.read().strip()

setup(
    version=version,
)



from flippy import recursive_map, recursive_filter, recursive_reduce, mem, flip
from flippy.distributions import Bernoulli, Categorical
from flippy.core import GlobalStore, ReadOnlyProxy
from flippy.core import SampleState, ReturnState
from flippy.interpreter import CPSInterpreter
from flippy.transforms import CPSTransform
from flippy.transforms import CPSFunction
from flippy.callentryexit import EnterCallState, ExitCallState, register_call_entryexit
from flippy.hashable import hashablelist, hashabledict
import ast
import math
import pytest
import traceback

from flippy.tests.test_transforms import trampoline

def geometric(p):
    x = Bernoulli(p).sample()
    if x == 0:
        return 0
    return 1 + geometric(p)

def algebra():
    def flip():
        return Bernoulli(0.5).sample()
    def NUM():
        return Categorical(range(5)).sample()
    def OP():
        return Categorical(['+', '*']).sample()
    def EQ():
        if flip():
            return NUM()
        else:
            return (NUM(), OP(), EQ())
    return EQ()

def check_trace(func, trace, *, args=(), kwargs={}, _emit_call_entryexit=False, return_value):
    ps = CPSInterpreter(_emit_call_entryexit=_emit_call_entryexit).initial_program_state(func)
    print(ast.unparse(CPSInterpreter().transform_from_func(func)))
    ps = ps.step(*args, **kwargs)

    for trace_idx, (dist, value) in enumerate(trace):
        if isinstance(ps, SampleState):
            assert ps.distribution.isclose(dist), (f'{trace_idx=}', ps)
            assert value in dist.support
            ps = ps.step(value)
        elif isinstance(ps, EnterCallState):
            assert dist == 'enter', (f'{trace_idx=}', ps, ps.function.__name__)
            assert ps.function.__name__ == value, (f'{trace_idx=}', ps, ps.function.__name__)
            ps = ps.step()
        elif isinstance(ps, ExitCallState):
            assert dist == 'exit', (f'{trace_idx=}', ps, ps.function.__name__)
            assert ps.function.__name__ == value, (f'{trace_idx=}', ps, ps.function.__name__)
            ps = ps.step()
        else:
            assert False, (f'{trace_idx=}', ps)

    assert isinstance(ps, ReturnState), ps
    assert ps.value == return_value

def test_interpreter():
    check_trace(geometric, [
        (Bernoulli(0.9), 0),
    ], args=(0.9,), return_value=0)

    check_trace(geometric, [
        (Bernoulli(0.8), 1),
        (Bernoulli(0.8), 1),
        (Bernoulli(0.8), 0),
    ], args=(0.8,), return_value=2)

    check_trace(algebra, [
        (Bernoulli(0.5), 1),
        (Categorical(range(5)), 3),
    ], return_value=3)

    check_trace(algebra, [
        (Bernoulli(0.5), 0),
        (Categorical(range(5)), 3),
        (Categorical(['+', '*']), '*'),
        # subtree on right
        (Bernoulli(0.5), 0),
        (Categorical(range(5)), 2),
        (Categorical(['+', '*']), '+'),
        (Bernoulli(0.5), 1),
        (Categorical(range(5)), 4),
    ], return_value=(3, '*', (2, '+', 4)))

def test_cps_map():
    def fn():
        def f(x):
            return Bernoulli(x).sample()
        return sum(recursive_map(f, [.1, .2]))

    check_trace(fn, [
        (Bernoulli(0.1), 1),
        (Bernoulli(0.2), 1),
    ], return_value=2)

def test_cps_filter():
    def fn():
        def f(x):
            return Categorical([1, 2, 3, 4]).sample()
        def is_even(x):
            return x % 2 == 0
        return sum(recursive_filter(is_even, recursive_map(f, [None] * 4)))

    check_trace(fn, [
        (Categorical([1, 2, 3, 4]), 1),
        (Categorical([1, 2, 3, 4]), 2),
        (Categorical([1, 2, 3, 4]), 3),
        (Categorical([1, 2, 3, 4]), 4),
    ], return_value=6)

def test_recursive_reduce():
    def fn():
        return recursive_reduce(lambda acc, x: acc + Bernoulli(x).sample(), [.1, .2], 0)

    check_trace(fn, [
        (Bernoulli(0.1), 1),
        (Bernoulli(0.2), 1),
    ], return_value=2)

def test_list_comprehension():
    # Simple case
    expected = [0, 1, 4]
    def fn():
        return [x**2 for x in range(3)]
    assert fn() == expected
    check_trace(fn, [], return_value=expected)

    # Multiple if statements.
    expected = [0, 2]
    def fn():
        return [x for x in range(5) if x % 2 == 0 if x < 3]
    assert fn() == expected
    check_trace(fn, [], return_value=expected)

    # Multiple generators
    expected = [(0, 0), (0, 1), (0, 2), (2, 0)]
    def fn():
        return [
            (x, y)
            for x in range(4)
            if x % 2 == 0
            for y in range(5)
            if x + y < 3
        ]
    assert fn() == expected
    check_trace(fn, [], return_value=expected)

    # Nested comprehensions
    expected = [[0], [0, 1], [0, 1, 2]]
    def fn():
        return [
            [y for y in range(x+1)]
            for x in range(3)
        ]
    assert fn() == expected
    check_trace(fn, [], return_value=expected)

    # Set comprehensions
    expected = {0, 1, 4, 9}
    def fn():
        return {
            x**2
            for x in [-3, -2, -1, 0, 1, 2, 3]
        }
    assert fn() == expected
    check_trace(fn, [], return_value=expected)

    # Dict comprehensions
    expected = {0: 0, 1: 1, 2: 4, 3: 9}
    def fn():
        return {
            x: x**2
            for x in range(4)
        }
    assert fn() == expected
    check_trace(fn, [], return_value=expected)

    # Checking something stochastic.
    def fn():
        return sum([Bernoulli(x).sample() for x in [.1, .2, .3]])
    check_trace(fn, [
        (Bernoulli(0.1), 1),
        (Bernoulli(0.2), 0),
        (Bernoulli(0.3), 1),
    ], return_value=2)

def test_check_exception():
    def test_fn():
        raise Exception('expected exception')
        return 3
    with pytest.raises(Exception) as e:
        check_trace(test_fn, [], return_value=3)
    # The right exception was raised.
    assert 'expected exception' in str(e)

    # Now check exception content to make sure it references code.
    _, exc, tb = e._excinfo
    # We show compiled code, which in this case isn't particularly readable.
    # This line will have to change depending on our compiled output.
    exception_line = 'raise _res_0'

    # First, we just check that the traceback will render the code.
    formatted = ''.join(traceback.format_exception(exc, None, tb))
    assert exception_line in formatted

    # This is a more detailed check of the contents, ensuring the filename is correct in traceback.
    last_entry = traceback.extract_tb(tb)[-1]
    assert 'test_fn' in last_entry.filename
    assert hex(id(test_fn)).removeprefix('0x') in last_entry.filename
    assert last_entry.line == exception_line

def test_control_flow_or():
    def fn_or():
        return Bernoulli(0.5).sample() or Bernoulli(0.5).sample()

    check_trace(fn_or, [
        (Bernoulli(0.5), 0),
        (Bernoulli(0.5), 1),
    ], return_value=1)

    check_trace(fn_or, [
        (Bernoulli(0.5), 1),
    ], return_value=1)

def test_control_flow_and():
    def fn_and():
        return Bernoulli(0.5).sample() and Bernoulli(0.5).sample()

    check_trace(fn_and, [
        (Bernoulli(0.5), 1),
        (Bernoulli(0.5), 1),
    ], return_value=1)

    check_trace(fn_and, [
        (Bernoulli(0.5), 0),
    ], return_value=0)

def test_control_flow_for_continue_break():
    def fn():
        rv = []
        for i in range(10):
            if i == 3:
                continue
            if i == 7:
                break
            rv += [i]
        return rv

    check_trace(fn, [], return_value=[0, 1, 2, 4, 5, 6])

def test_conditional_reassign():
    def fn():
        x = 0
        if Bernoulli(0.5).sample():
            x = 42
        return x

    check_trace(fn, [(Bernoulli(0.5), 0)], return_value=0)
    check_trace(fn, [(Bernoulli(0.5), 1)], return_value=42)

def test_load_then_store_in_new_scope():
    def fn():
        x = 0
        z = Bernoulli(0.5).sample()
        x = x + z
        return x

    check_trace(fn, [(Bernoulli(0.5), 0)], return_value=0)
    check_trace(fn, [(Bernoulli(0.5), 1)], return_value=1)

def test_conditionally_defined():
    # This can break implementations of scope passing that assume
    # variable lists can be fully statically determined.
    def fn():
        z = Bernoulli(0.5).sample()

        if z:
            y = 42

        if z:
            return y
        else:
            return -1

    check_trace(fn, [(Bernoulli(0.5), 0)], return_value=-1)
    check_trace(fn, [(Bernoulli(0.5), 1)], return_value=42)

def test_closure_issues():
    # Making sure we throw for important closure issues.

    # If incorrect, this returns 'bad'.
    with pytest.raises(SyntaxError) as err:
        def fn():
            def nested():
                return x
            x = 'bad'
            Bernoulli(0.5).sample() # Arbitrary. To put rest of function in continuation.
            x = 'good'
            return nested()
        check_trace(fn, [(Bernoulli(0.5), 0)], return_value='good')
    assert 'must be immutable' in str(err)

    # If incorrect, this fails to run with NameError.
    with pytest.raises(SyntaxError) as err:
        def fn():
            def nested():
                return x
            Bernoulli(0.5).sample() # Arbitrary. To put rest of function in continuation.
            x = 'good'
            return nested()
        check_trace(fn, [(Bernoulli(0.5), 0)], return_value='good')
    assert 'must be defined before' in str(err)

def test_global_store_proxy():
    global_store = None
    def f():
        return global_store.get('a', 'no value')

    cps = CPSInterpreter()
    code = cps.transform_from_func(f)
    context = {
        "_cps": cps,
        'global_store': cps.global_store_proxy,
        "CPSFunction": CPSFunction,
        "hashablelist": hashablelist,
        "hashabledict": hashabledict,
    }
    exec(ast.unparse(code), context)
    f = context['f']

    with cps.set_global_store(GlobalStore()):
        assert trampoline(f()) == 'no value'

    with cps.set_global_store(GlobalStore({'a': 100})):
        assert trampoline(f()) == 100

proxy_forking_bags = Categorical(['bag0', 'bag1', 'bag2'])
def proxy_forking_value(_bag):
    return Categorical(range(10)).sample()
proxy_forking_value = CPSInterpreter().non_cps_callable_to_cps_callable(proxy_forking_value)
def proxy_forking():
    value = mem(proxy_forking_value)
    return [
        value(proxy_forking_bags.sample())
        for _ in range(5)
    ]

def test_global_store_proxy_forking():
    cps = CPSInterpreter()
    s0 = cps.initial_program_state(proxy_forking).step()
    assert s0.init_global_store.store == {}

    s1a = s0.step('bag0').step(3)
    assert s0.init_global_store.store == {}, 'Make sure original state was not modified'
    assert s1a.init_global_store.store == {(proxy_forking_value, ('bag0',), ()): 3}

    # A sanity check, we shouldn't need to resample for bag0 now.
    s = s1a.step('bag0')
    assert isinstance(s, SampleState) and s.distribution == proxy_forking_bags

    # Now, we test forking by restarting at s0
    s1b = s0.step('bag1').step(7)
    assert s0.init_global_store.store == {}, 'Make sure original state was not modified'
    assert s1a.init_global_store.store == {(proxy_forking_value, ('bag0',), ()): 3}, 'Make sure sibling state was not modified'
    assert s1b.init_global_store.store == {(proxy_forking_value, ('bag1',), ()): 7}

    # This forked state should not pick up on state from s1a
    s2b = s1b.step('bag0').step(5)
    assert s0.init_global_store.store == {}, 'Make sure original state was not modified'
    assert s1a.init_global_store.store == {(proxy_forking_value, ('bag0',), ()): 3}, 'Make sure sibling state was not modified'
    assert s1b.init_global_store.store == {(proxy_forking_value, ('bag1',), ()): 7}, 'Make sure original state was not modified'
    assert s2b.init_global_store.store == {(proxy_forking_value, ('bag1',), ()): 7, (proxy_forking_value, ('bag0',), ()): 5}, 'Make sure original state was not modified'

def test_Distribution_generic_methods():
    # other than observe and sample, Distribution methods
    # should be run deterministically
    def f():
        x = Bernoulli(0.5)
        return x.log_probability(1)
    check_trace(
        f, [], return_value=math.log(.5)
    )

def test_global_store_programstate_hashing():
    def g_base(i):
        return 100
    g_base = CPSInterpreter().non_cps_callable_to_cps_callable(g_base)
    g = mem(g_base)

    def f_global_diff():
        i = flip(.2, name='i')
        j = g(i)
        return j

    def f_global_same():
        i = flip(.2, name='i')
        j = g(0)
        return j

    def f_noglobal():
        i = flip(.2, name='i')
        j = g_base(i)
        return j

    ps = CPSInterpreter().initial_program_state(f_global_diff)
    i_ps = ps.step()
    assert i_ps.step(1) != i_ps.step(0)
    assert (g_base, (True, ), ()) in i_ps.step(1).init_global_store.store
    assert (g_base, (False, ), ()) in i_ps.step(0).init_global_store.store
    assert i_ps.step(1).value == i_ps.step(0).value == 100

    ps = CPSInterpreter().initial_program_state(f_global_same)
    i_ps = ps.step()
    assert i_ps.step(1) == i_ps.step(0)
    assert (g_base, (False, ), ()) in i_ps.step(0).init_global_store.store
    assert (g_base, (False, ), ()) in i_ps.step(1).init_global_store.store
    assert i_ps.step(1).value == i_ps.step(0).value == 100

    ps = CPSInterpreter().initial_program_state(f_noglobal)
    i_ps = ps.step()
    assert i_ps.step(1) == i_ps.step(0)
    assert i_ps.step(0).init_global_store.store == i_ps.step(1).init_global_store.store == {}
    assert i_ps.step(1).value == i_ps.step(0).value == 100

def test_deterministic_nested_call_entryexit():
    def model():
        @register_call_entryexit
        def f1():
            return 'a'

        @register_call_entryexit
        def f2():
            return 'b' + f1()

        @register_call_entryexit
        def f3():
            return 'c' + f2()
        return f3()
    ps = CPSInterpreter().initial_program_state(model)
    ps_seq = []
    while not isinstance(ps, ReturnState):
        ps = ps.step()
        ps_seq.append(ps)
    assert isinstance(ps_seq[0], EnterCallState)
    assert ps_seq[0].function.__name__ == 'f3'
    assert isinstance(ps_seq[1], EnterCallState)
    assert ps_seq[1].function.__name__ == 'f2'
    assert isinstance(ps_seq[2], EnterCallState)
    assert ps_seq[2].function.__name__ == 'f1'
    assert isinstance(ps_seq[3], ExitCallState)
    assert ps_seq[3].function.__name__ == 'f1'
    assert isinstance(ps_seq[4], ExitCallState)
    assert ps_seq[4].function.__name__ == 'f2'
    assert isinstance(ps_seq[5], ExitCallState)
    assert ps_seq[5].function.__name__ == 'f3'

    assert ps.value == 'cba'

def test_call_entry_exit_cps_function_requirement():
    with pytest.raises(AssertionError) as e:
        @register_call_entryexit
        def f1():
            return 'a'
    assert 'can only be applied to transformed functions' in str(e)

def test_program_state_identity_with_closures():
    def f():
        i = flip(.65, name='i')
        def g():
            j = flip(.23, name='j')
            return j + i
        return g()

    ps = CPSInterpreter().initial_program_state(f)
    ps = ps.step()

    # identity of program states depends on g() and its closure (which includes i)
    ps_0a = ps.step(0)
    ps_0b = ps.step(0)
    ps_1 = ps.step(1)
    assert ps_0a == ps_0b
    assert hash(ps_0a) == hash(ps_0b)
    assert id(ps_0a) != id(ps_0b)
    assert ps_0a != ps_1
    assert hash(ps_0a) != hash(ps_1)

    # Names are the same regardless of locals
    assert ps_0a.name == ps_0b.name
    assert ps_0a.name == ps_1.name

    # g in each thread are equal, but generally not identical objects
    assert ps_0a.stack[1].locals['g'] == ps_0b.stack[1].locals['g']
    assert id(ps_0a.stack[1].locals['g']) != id(ps_0b.stack[1].locals['g'])
    # try:
    #     # its possible but unlikely that they are identical due to memory reuse
    #     assert id(ps.step(0).stack[1].locals['g']) != id(ps.step(0).stack[1].locals['g'])
    # except AssertionError:
    #     assert id(ps.step(0).stack[1].locals['g']) != id(ps.step(0).stack[1].locals['g'])

def _assert_stack_frame_fn_match(stack, fn_names):
    assert len(stack.stack_frames) == len(fn_names)
    for frame, fn_name in zip(stack.stack_frames, fn_names):
        if fn_name == '<root>':
            prefix = fn_name
        else:
            prefix = f'def {fn_name}('
        assert frame.func_src.startswith(prefix), (prefix, frame.func_src[:100])

def test_program_state_identity_with_loops():
    def f():
        ct = 0
        for _ in range(3):
            ct += Bernoulli(0.5).sample()
        return ct

    ps = CPSInterpreter().initial_program_state(f)
    ps = ps.step()
    ps0 = ps.step(0)
    ps00 = ps0.step(0)
    assert ps0 != ps00
    assert hash(ps0) != hash(ps00)
    assert ps0.name != ps00.name
    _assert_stack_frame_fn_match(ps0.name, ['<root>', 'f', '_loop_fn_1', '_loop_fn_1'])
    _assert_stack_frame_fn_match(ps00.name, ['<root>', 'f', '_loop_fn_1', '_loop_fn_1', '_loop_fn_1'])

def test_program_state_identity_with_loop_continue():
    def f():
        ct = 0
        for i in range(3):
            if Bernoulli(0.5).sample():
                ct += i * 2
                continue
            ct += i
        return ct

    ps = CPSInterpreter().initial_program_state(f)
    ps = ps.step()
    ps0 = ps.step(0)
    ps00 = ps0.step(0)
    ps1 = ps.step(1)
    ps10 = ps1.step(0)
    assert ps00 != ps10
    assert hash(ps00) != hash(ps10)
    # TODO FIX THIS
    with pytest.raises(AssertionError):
        assert ps00.name != ps10.name
        # HACK: make this check for line numbers
        _assert_stack_frame_fn_match(ps0.name, ['<root>', 'f', '_loop_fn_5', '_loop_fn_5'])
        _assert_stack_frame_fn_match(ps00.name, ['<root>', 'f', '_loop_fn_5', '_loop_fn_5', '_loop_fn_5'])

def test_call_entryexit_skipping():
    def model():
        @register_call_entryexit
        def f(p):
            return flip(p) + flip(p)
        return f(.6) + 20

    ps = CPSInterpreter().initial_program_state(model)
    enter_ps = ps.step()

    # Don't run the model and return 100
    exit_ps = enter_ps.step(False, 100)
    assert exit_ps.value == 100
    return_ps = exit_ps.step()
    assert return_ps.value == 120

    # Run the function and set bernoulli's to 1
    exit_ps = enter_ps.step(True).step(1).step(1)
    assert exit_ps.value == 2
    return_ps = exit_ps.step()
    assert return_ps.value == 22

def test_decorated_recursive_functions():
    def decorator(fn):
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)
        return wrapper

    def f():
        @decorator
        def wrapped():
            inner_ref = wrapped
            return inner_ref
        outer_ref = wrapped
        inner_ref = wrapped()
        assert outer_ref is inner_ref
        return 1

    ps = CPSInterpreter().initial_program_state(f)
    ps.step()

def test_program_state_stack_name_with_local_change():
    def f(i):
        x = Bernoulli(0.5).sample()
        return x

    ps = CPSInterpreter().initial_program_state(f)
    assert ps.step(0).name == ps.step(1).name

def test_sampling_in_instance_method_call():
    class C:
        def flip(self, p):
            return Bernoulli(p).sample()

    def call_as_instance_attribute():
        c = C()
        return c.flip(.63)

    def call_as_class_attribute():
        c = C()
        return C.flip(c, .63)

    def tests(f):
        # Execute to the sample statement in the class method
        in_class_flip_ps = CPSInterpreter().initial_program_state(f).step()
        assert isinstance(in_class_flip_ps, SampleState)
        assert in_class_flip_ps.distribution.isclose(Bernoulli(.63))

        # Check that instance is referenced in transformed method
        class_flip_locals = in_class_flip_ps.stack[-1].locals
        assert isinstance(class_flip_locals['self'], C)

        # Check instance properties in main function scope
        f_class_locals = in_class_flip_ps.stack[-2].locals
        assert class_flip_locals['self'] is f_class_locals['c']
        assert isinstance(f_class_locals['c'], C)
        assert f_class_locals['c'].flip.__qualname__[-len("C.flip"):] == "C.flip"

    tests(call_as_instance_attribute)
    tests(call_as_class_attribute)

def test_transformed_class_method_does_not_shadow():
    def func(p):
        return 100
    nonlocal_func = func

    # Test instance method
    class C:
        def func(self):
            assert func is nonlocal_func, "nonlocal function has been redefined"
            return 101

    def f():
        c = C()
        return c.func()

    assert f() == 101
    assert CPSInterpreter().initial_program_state(f).step().value == 101

    # Now test class method
    class C:
        @classmethod
        def func(cls):
            assert func is nonlocal_func, "nonlocal function has been redefined"
            return 101

    def f_class():
        return C.func()

    assert f_class() == 101
    assert CPSInterpreter().initial_program_state(f_class).step().value == 101

    # Now test static method
    class C:
        @staticmethod
        def func():
            assert func is nonlocal_func, "nonlocal function has been redefined"
            return 101

    def f_static():
        return C.func()

    assert f_static() == 101
    assert CPSInterpreter().initial_program_state(f_static).step().value == 101

    # Now test nested class method
    class C1:
        class C2:
            def func(self):
                assert func is nonlocal_func, "nonlocal function has been redefined"
                return 101

    def f_nested():
        c1 = C1()
        c2 = c1.C2()
        return c2.func()

    assert CPSInterpreter().initial_program_state(f_nested).step().value == 101

def test_calling_transformed_staticmethod():
    class C:
        @staticmethod
        def flip():
            p = .66
            return Bernoulli(p).sample()

        @staticmethod
        def flip2(p):
            return Bernoulli(p).sample()

        def flip_no_decorator():
            p = .66
            return Bernoulli(p).sample()

    def call_from_class():
        return C.flip()

    def call_from_class_with_arg():
        return C.flip2(.66)

    def call_from_class_no_decorator():
        return C.flip_no_decorator()

    def call_from_instance():
        c = C()
        return c.flip()

    def call_from_instance_no_decorator():
        c = C()
        return c.flip_no_decorator()

    ps = CPSInterpreter().initial_program_state(call_from_class).step()
    assert ps.distribution.isclose(Bernoulli(.66))

    ps = CPSInterpreter().initial_program_state(call_from_class_with_arg).step()
    assert ps.distribution.isclose(Bernoulli(.66))

    ps = CPSInterpreter().initial_program_state(call_from_class_no_decorator).step()
    assert ps.distribution.isclose(Bernoulli(.66))

    ps = CPSInterpreter().initial_program_state(call_from_instance).step()
    assert ps.distribution.isclose(Bernoulli(.66))

    with pytest.raises(TypeError) as e:
        ps = CPSInterpreter().initial_program_state(call_from_instance_no_decorator).step()
        assert isinstance(e.value, TypeError) and "0 positional argument" in str(e.value)

def test_calling_transformed_classmethod():
    class C:
        p = .66
        def __init__(self):
            self.p = .22
        @classmethod
        def flip(cls):
            return Bernoulli(cls.p).sample()

    def call_from_instance():
        c = C()
        return c.flip()

    def call_from_class():
        return C.flip()

    ps = CPSInterpreter().initial_program_state(call_from_instance).step()
    assert ps.distribution.isclose(Bernoulli(.66))

    ps = CPSInterpreter().initial_program_state(call_from_class).step()
    assert ps.distribution.isclose(Bernoulli(.66))

def test_generate_unique_method_name():
    def func(): pass

    class C:
        def func(self): pass
        class NC: #nested class
            def func(self): pass
    c = C()
    nc = C.NC()

    def class_factory():
        class LC: #locally defined class
            def func(self): pass
        return LC

    LC = class_factory()
    lc = LC()

    funcs_names = [
        (func, None),
        (C.func, '__C_func'),
        (c.func, '__C_func'),
        (C.NC.func, "__C_NC_func"),
        (nc.func, "__C_NC_func"),
        (LC.func, '__LC_func'),
        (lc.func, '__LC_func')
    ]
    for f, n in funcs_names:
        gen_name = CPSInterpreter().generate_unique_method_name(f)
        assert n == gen_name, (f, n, gen_name)

def test_CPSFunction_hashing_and_equality():
    # We assume functions are equal up to source code and closure
    # This doesn't include global scope since that's handled by program state
    def model():
        i = Bernoulli(0.5).sample() + Bernoulli(0.5).sample()
        def f():
            return i + 1
        return f

    ps = CPSInterpreter().initial_program_state(model)
    fs = [ps.step().step(b1).step(b2).value for b1 in [0, 1] for b2 in [0, 1]]
    assert all(isinstance(f, CPSFunction) for f in fs), "All functions should be CPS transformed"
    assert len(set(fs)) == 3, "Functions are unique up to their source code and closures"
    assert len(set(f.func_src for f in fs)) == 1
    assert len(set(f.closure for f in fs)) == 3
    assert set(f.closure['i'] for f in fs) == {0, 1, 2}

def test_CPSFunction_closure_for_dynamically_transformed_functions():
    # This tests whether functions that are transformed at runtime
    # capture closure variables correctly.
    def func_factory(i, _cont, **kws):
        def func():
            return i
        return lambda : _cont(func)
    # Don't transform func_factory so we can dynamically transform func at runtime
    setattr(func_factory, CPSTransform.is_transformed_property, True)

    def model():
        i = Bernoulli(0.5).sample()
        fn = func_factory(i)
        return fn()

    ps = CPSInterpreter(_emit_call_entryexit=True).initial_program_state(model)
    func_closure = ps.step().step().step(0).step().step().function.closure
    assert func_closure.get('i') == 0

def test_callsite_addressing_on_single_line():
    def m():
        def f(i):
            return 1
        return f(f(f(1)))

    ps = CPSInterpreter(_emit_call_entryexit=True).initial_program_state(m)
    ps = ps.step().step()
    call1 : EnterCallState = ps
    ps = ps.step().step()
    call2 : EnterCallState = ps
    ps = ps.step().step()
    call3 : EnterCallState = ps

    # Address components we expect to stay the same for different calls
    call_names = (call1.function.__name__, call2.function.__name__, call3.function.__name__)
    assert set(call_names) == {"f"}, "Call function names should be the same"
    call_linenos = (call1.stack[-1].lineno, call2.stack[-1].lineno, call3.stack[-1].lineno)
    assert len(set(call_linenos)) == 1, "Call line numbers should be the same"
    call_substacks = (call1.stack[:-1], call2.stack[:-1], call3.stack[:-1])
    assert len(set(call_substacks)) == 1, "Call sub stacks should be the same"

    # Address components we expect to be distinct for different calls
    call_ids = (call1.stack[-1].call_id, call2.stack[-1].call_id, call3.stack[-1].call_id)
    assert len(set(call_ids)) == 3, "Call IDs should be unique"

    # Thus the call stacks should be distinct
    call_stacks = (call1.stack, call2.stack, call3.stack)
    assert len(set(call_stacks)) == 3, "Call stacks should be the same"

def test_cps_loop_entryexit():
    def binom(k, p):
        ct = 0
        for i in range(k):
            ct += Bernoulli(p).sample()
        return ct
    check_trace(binom, [
        ('enter', 'binom'),
        ('enter', '_loop_fn_1'),
        (Bernoulli(0.5), 0),
        ('enter', '_loop_fn_1'),
        (Bernoulli(0.5), 1),
        ('enter', '_loop_fn_1'),
        (Bernoulli(0.5), 0),
        ('enter', '_loop_fn_1'),
        ('exit', '_loop_fn_1'),
        ('exit', '_loop_fn_1'),
        ('exit', '_loop_fn_1'),
        ('exit', '_loop_fn_1'),
        ('exit', 'binom'),
    ], args=(3, 0.5), return_value=1, _emit_call_entryexit=True)

def test_modifying_global_variable():
    global data
    data = [1, 2, 3]

    def f():
        return sum(data)
    ps = CPSInterpreter().initial_program_state(f)

    ret_val = ps.step().value
    assert ret_val == 6

    data = [1, 2, 3, 4]
    ret_val = ps.step().value
    assert ret_val == 10

def test_modifying_closure_variable():
    data = [1, 2, 3]

    def f():
        return sum(data)
    ps = CPSInterpreter().initial_program_state(f)

    ret_val = ps.step().value
    assert ret_val == 6

    data = [1, 2, 3, 4]
    ret_val = ps.step().value

    # This is inconsistent with normal Python closure semantics
    with pytest.raises(AssertionError):
        assert ret_val == 10



import dataclasses
from functools import cached_property
from typing import Mapping, Hashable, Callable, Any, List, Union, Dict, Tuple

from flippy.inference.mcmc.trace import Trace

@dataclasses.dataclass
class MCMCDiagnosticsEntry:
    old_trace : Trace
    new_trace : Trace
    log_acceptance_threshold : float
    log_acceptance_ratio : float
    sampled_trace : Trace
    accept : bool
    save_sample : bool
    auxiliary_vars : Any

@dataclasses.dataclass
class MCMCDiagnostics:
    history : List[MCMCDiagnosticsEntry] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        self.shortened_names = {}

    def calc_variable_values(self, only_sampled : bool = True):
        variable_values = {}

        if only_sampled:
            history = self.sampled_history
        else:
            history = self.history
        for i, s in enumerate(history):
            if not s.save_sample:
                continue
            for e in s.sampled_trace.entries():
                if not e.is_sample:
                    continue
                if not isinstance(e.name, str) and e.name[0] == (None, None): #HACK
                    name = self.shorten_name(e.name)
                else:
                    name = e.name
                if name not in variable_values:
                    variable_values[name] = [None]*len(history)
                variable_values[name][i] = e.value
        return variable_values

    def append(self, entry : MCMCDiagnosticsEntry):
        self.history.append(entry)

    @cached_property
    def sampled_history(self) -> List[MCMCDiagnosticsEntry]:
        return [e for e in self.history if e.save_sample]

    def shorten_name(self, name):
        if name in self.shortened_names:
            return self.shortened_names[name]
        top_stack_frame = name[-1]
        short_name = ShortVariableName(
            len(name),
            get_function_name(top_stack_frame[0]),
            top_stack_frame[1]
        )
        if short_name in self.shortened_names.values():
            short_name = short_name.increment_version()
            while short_name in self.shortened_names.values():
                short_name = short_name.increment_version()
        self.shortened_names[name] = short_name
        return short_name

    def plot_parameter_samples(
        self,
        top_n_vars = None,
        subplot_height = 8,
        subplot_width = 2,
        sort_varnames = False,
        only_sampled = True,
    ):
        import pandas as pd
        import matplotlib.pyplot as plt
        import numpy as np
        if top_n_vars in (None, -1):
            top_n_vars = len(self.calc_variable_values(only_sampled=only_sampled))

        df = pd.DataFrame(self.calc_variable_values(only_sampled=only_sampled))
        isna = df.isna().sum(axis=0)
        if sort_varnames:
            varnames = sorted(isna.index, key=lambda x: isna[x])
        else:
            varnames = isna.index
        fig, axes = plt.subplots(top_n_vars, 1, figsize=(subplot_height, subplot_width*top_n_vars))
        for i in range(top_n_vars):
            var_vals = df[varnames[i]].values
            ax = axes[i]
            if all(isinstance(v, (tuple, list)) for v in var_vals if v is not None):
                var_vals = [list(v) if v is not None else None for v in var_vals]
                max_row = max(len(v) for v in var_vals if v is not None)
                var_vals = [v if v is not None else [None]*max_row for v in var_vals]
                var_vals = np.array(var_vals)
            elif all(isinstance(v, str) for v in var_vals if v is not None):
                all_values = sorted(set([v for v in var_vals if v is not None]))
                one_hot = np.zeros((len(var_vals), len(all_values)))
                for j, v in enumerate(var_vals):
                    if v is not None:
                        one_hot[j, all_values.index(v)] = 1
                    else:
                        one_hot[j, :] = np.nan
                var_vals = one_hot
            else:
                var_vals = np.array(var_vals)
            ax.plot(var_vals, 'x-')
            ax.set_title(str(varnames[i]))
            ax.set_xlim(0, len(var_vals))
        plt.tight_layout()
        return fig

    def plot_sampler_stats(self):
        import pandas as pd
        import matplotlib.pyplot as plt
        import numpy as np
        trace_scores = []
        for e in self.history:
            trace_scores.append(dict(
                accept = e.accept,
                log_acceptance_threshold = e.log_acceptance_threshold,
                log_acceptance_ratio = e.log_acceptance_ratio,
                old_trace = e.old_trace.total_score,
                new_trace = e.new_trace.total_score,
                sampled_trace = e.sampled_trace.total_score,
            ))
        trace_scores = pd.DataFrame(trace_scores)

        fig, axes = plt.subplots(2, 1, figsize=(8, 5))
        trace_scores['accept'].expanding().mean().plot(ax=axes[0])
        axes[0].set_title(f'Cumulative acceptance rate (final: {trace_scores["accept"].mean():.2f})')
        axes[0].set_ylim(0, 1)
        axes[0].set_xlim(-10, len(trace_scores))
        trace_scores['sampled_trace'].plot(ax=axes[1])
        axes[1].plot(
            [r['sampled_trace'] if r['accept'] else None for _, r in trace_scores.iterrows()],
            'xk', markersize=2
        )
        axes[1].set_title('Accepted trace scores')
        axes[1].set_xlim(-10, len(trace_scores))
        plt.tight_layout()

def get_function_name(fn_src):
    lines = fn_src.split('\n')
    for line in lines:
        if 'def ' in line:
            return line.split('def ')[1].split('(')[0]

@dataclasses.dataclass(frozen=True)
class ShortVariableName:
    depth : int
    function_name : str
    lineno : int
    version : int = 0

    def increment_version(self):
        return dataclasses.replace(self, version=self.version+1)

    def __str__(self):
        return f"{self.function_name}.{self.depth}.{self.lineno}.{self.version}"



from flippy import infer, flip, keep_deterministic, Bernoulli, mem, factor
from flippy.map import independent_map
from flippy.inference.simpleenumeration import SimpleEnumeration
from flippy.inference.enumeration import Enumeration

def test_independent_map():
    iterator = (.1, .2, .3, .4, .5, .6, .7, .8, .9)
    def f1():
        def g(i):
            return flip(i) + 1
        x = independent_map(g, iterator)
        return x

    def f2():
        def g(i):
            return flip(i) + 1
        x = tuple([g(i) for i in iterator])
        return x

    res1 = SimpleEnumeration(f1).run()
    res2 = SimpleEnumeration(f2).run()
    res3 = Enumeration(f1).run()
    assert res1.isclose(res2)
    assert res1.isclose(res3)

def test_independent_map_with_mem():
    def f(n):
        @mem
        def h(i):
            return flip(0.5)
        def g(p):
            i = flip(p) + flip(p)
            return h(i)
        return independent_map(g, tuple([i/(n - 1) for i in range(n)]))

    e_res = SimpleEnumeration(f).run(n=3)
    ge_res = Enumeration(f).run(n=3)
    assert e_res.isclose(ge_res)



"""
This module implements a `ProgramState` API for calling and returning from
function calls.
"""

from typing import Callable, TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from flippy.core import ProgramState
from flippy.types import CPSCallable, Continuation, Thunk
from flippy.transforms import CPSTransform, CPSFunction
from flippy.hashable import hashabledict

if TYPE_CHECKING:
    from flippy.interpreter import CPSInterpreter, Stack

class EnterCallState(ProgramState):
    """
    Represents the state of a program when entering a function call.
    This can be used to intercept function calls (e.g., for callsite caching;
    see [Ritchie, Stuhlmüller & Goodman (2016)](https://proceedings.mlr.press/v51/ritchie16.html)).
    """
    def __init__(
        self,
        f: CPSFunction,
        args: Tuple,
        kwargs: Dict,
        continuation: Callable[[bool, Any], Union[Thunk, ProgramState]]=None,
        cps: 'CPSInterpreter'=None,
        stack: 'Stack'=None,
    ):
        super().__init__(
            continuation=continuation,
            name=(f, args, hashabledict(kwargs)),
            cps=cps,
            stack=stack,
        )
        self.function = f
        self.args = args
        self.kwargs = kwargs

    @property
    def is_root_call(self) -> bool:
        return len(self.stack) == 1

    def skip(self, value) -> 'ExitCallState':
        return self.step(False, value)

    def step(self, run_func: bool = True, res: Any = None) -> 'ProgramState':
        return ProgramState.step(self, run_func, res)

class ExitCallState(ProgramState):
    """Represents the state of a program when exiting a function call."""
    def __init__(
        self,
        f: CPSFunction,
        args: Tuple,
        kwargs: Dict,
        value: Any,
        continuation: Continuation=None,
        cps: 'CPSInterpreter'=None,
        stack: 'Stack'=None,
    ):
        super().__init__(
            continuation=continuation,
            name=(f, args, hashabledict(kwargs), value),
            cps=cps,
            stack=stack,
        )
        self.function = f
        self.args = args
        self.kwargs = kwargs
        self.value = value

def enter_call_event(
    f: CPSFunction,
    args: Tuple,
    kwargs: Dict,
    _cont: Continuation=None,
    _cps: 'CPSInterpreter'=None,
    _stack: 'Stack'=None,
):
    return EnterCallState(
        f=f,
        args=args,
        kwargs=kwargs,
        continuation=lambda run_func=True, res=None: _cont((run_func, res)),
        cps=_cps,
        stack=_stack,
    )
setattr(enter_call_event, CPSTransform.is_transformed_property, True)

def exit_call_event(
    f: CPSFunction,
    args: Tuple,
    kwargs: Dict,
    value: Any,
    _cont: Continuation=None,
    _cps: 'CPSInterpreter'=None,
    _stack: 'Stack'=None,
):
    return ExitCallState(
        f=f,
        args=args,
        kwargs=kwargs,
        value=value,
        continuation=lambda : _cont(None),
        cps=_cps,
        stack=_stack,
    )
setattr(exit_call_event, CPSTransform.is_transformed_property, True)

def register_call_entryexit(f: CPSCallable) -> CPSCallable:
    assert isinstance(f, CPSFunction), \
        f'Error registering {f.__name__}: `register_call_entryexit` can only be applied to transformed functions'
    def call_entryexit_wrapper(*args, **kwargs):
        kwargs = hashabledict(kwargs)
        run_func, res = enter_call_event(f, args, kwargs)
        if run_func:
            res = f(*args, **kwargs)
        exit_call_event(f, args, kwargs, res)
        return res
    return call_entryexit_wrapper



import math
from collections import OrderedDict
from typing import Generic, TypeVar

ISCLOSE_RTOL = 1e-5
ISCLOSE_ATOL = 1e-8

def isclose(a, b, *, rtol=ISCLOSE_RTOL, atol=ISCLOSE_ATOL):
    return math.isclose(a, b, rel_tol=rtol, abs_tol=atol)

def logsumexp(*args):
    max_arg = max(args)
    return max_arg + math.log(sum(math.exp(arg - max_arg) for arg in args))

def softmax_dict(d: dict):
    max_value = max(d.values())
    keys, values = zip(*d.items())
    exp_values = [math.exp(v - max_value) for v in values]
    total = sum(exp_values)
    return {k: v/total for k, v in zip(keys, exp_values)}

KT = TypeVar('KT')
VT = TypeVar('VT')

class LRUCache(Generic[KT, VT]):
    def __init__(self, max_size=1024):
        self.max_size = max_size if max_size is not None else float('inf')
        self._cache = OrderedDict()
        self.hits = 0
        self.misses = 0

    def clear(self):
        self._cache.clear()

    def __getitem__(self, key):
        val = self._cache.pop(key)
        self._cache[key] = val
        return val

    def __setitem__(self, key, val):
        if key in self._cache:
            self._cache.pop(key)
        self._cache[key] = val
        if len(self._cache) > self.max_size:
            self._cache.popitem(last=False)
        self._cache[key] = val

    def __contains__(self, key: KT) -> bool:
        contains = key in self._cache
        if contains:
            self.hits += 1
        else:
            self.misses += 1
        return contains

    def keys(self):
        return self._cache.keys()

    def __len__(self) -> int:
        return len(self._cache)

    def __repr__(self) -> str:
        return f"LRUCache({self._cache}, max_size={self.max_size})"

class PackagePlaceholder:
    def __init__(self, package_name: str):
        self.package_name = package_name

    def __getattr__(self, item):
        raise ImportError(f"Package '{self.package_name}' is not installed. "
                          f"Please install it to use this feature.")

    def __repr__(self):
        return f"PackagePlaceholder({self.package_name})"



from flippy import infer, flip, mem, condition, draw_from, uniform, normal
from flippy.distributions.builtin_dists import Bernoulli, Categorical, Multinomial, \
    Gaussian, Normal, Gamma, Uniform, Beta, Binomial, Geometric, Poisson, \
    BetaBinomial, Dirichlet, DirichletMultinomial, Mixture
from flippy.tools import isclose

def test_distributions():
    flip()
    draw_from(3)
    uniform()
    normal()

    for dist in [
        Bernoulli(0.5), Uniform(0, 1), Beta(2, 5), Binomial(trials=10, p=0.5),
        Categorical((1, 2, 3)),
        Multinomial(
            categorical_support=("A", "B", "C"),
            trials=10,
            probabilities=(0.2, 0.5, 0.3)
        ),
        Gaussian(0, 1), Normal(0, 1), Gamma(2, 2), Geometric(0.5),
        Poisson(rate=3), BetaBinomial(trials=2, alpha=1.2, beta=3.3),
        Dirichlet([0.2, 0.5, 0.3]),
        DirichletMultinomial(
            trials=10,
            alphas=[0.2, 0.5, 0.3]
        ),
        Mixture(
            distributions=[
                Bernoulli(0.5),
                Gaussian(0, 1),
            ],
            weights=[0.5, 0.5]
        )
    ]:
        print(dist)
        x = dist.sample()
        assert dist.prob(x) > 0

def test_Enumeration():
    @infer(method="Enumeration")
    def simple_model():
        x = flip(0.5)
        y = flip(0.5)
        condition(x >= y)
        return x + y

    result = simple_model()
    assert dict(result) == {2: 1/3, 1: 1/3, 0: 1/3}

def test_MetropolisHastings():
    @infer(method="MetropolisHastings", samples=1000)
    def simple_model_mh():
        p = uniform()
        x = flip(p)
        y = flip(p)
        condition(x >= y)
        return x + y
    result_mh = simple_model_mh()

def test_mem():
    @mem
    def g(i):
        return flip()

    @infer
    def mem_model():
        return g(1) + g(1)

    result_mem = mem_model()
    assert dict(result_mem) == {2: 1/2, 0: 1/2}

def test_likelihood_weighting():
    @infer(method="LikelihoodWeighting", samples=1000, seed=13842)
    def model(p):
        def geometric():
            x = Bernoulli(p).sample()
            if x == 1:
                return 1
            return 1 + geometric()
        return geometric()

    param = 0.98
    expected = 1/param

    lw_dist = model(param)
    assert isclose(expected, lw_dist.expected_value(), atol=1e-2), 'Should be somewhat close to expected value'



from collections import defaultdict
from typing import Generic

from flippy.core import ProgramState, ReturnState, SampleState, ObserveState, InitialState
from flippy.interpreter import CPSInterpreter
from flippy.distributions import Categorical, RandomNumberGenerator
from flippy.types import Element
from flippy.inference.inference import InferenceAlgorithm

class SamplePrior(InferenceAlgorithm[Element]):
    """Sample from the prior and ignore observation statements"""
    def __init__(self, function, samples : int, seed=None):
        self.function = function
        self.seed = seed
        self.samples = samples

    @property
    def is_cachable(self):
        return self.seed is not None

    def run(self, *args, **kws) -> Categorical[Element]:
        rng = RandomNumberGenerator(self.seed)
        return_counts = defaultdict(int)
        init_ps = CPSInterpreter().initial_program_state(self.function)
        for _ in range(self.samples):
            ps = init_ps.step(*args, **kws)
            while not isinstance(ps, ReturnState):
                if isinstance(ps, SampleState):
                    value = ps.distribution.sample(rng=rng)
                    ps = ps.step(value)
                elif isinstance(ps, ObserveState):
                    ps = ps.step()
                else:
                    raise ValueError("Unrecognized program state message")
            return_counts[ps.value] += 1
        total_prob = sum(return_counts.values())
        return_probs = {e: p/total_prob for e, p in return_counts.items()}
        return Categorical.from_dict(return_probs)
